# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Base module variables."""
from ._version import get_versions

__version__ = get_versions()["version"]
del get_versions

__org__ = "nipreps"
__packagename__ = "nibabies"
__copyright__ = "Copyright 2021, Center for Reproducible Neuroscience, Stanford University"
__credits__ = (
    "Contributors: please check the ``.zenodo.json`` file at the top-level folder"
    "of the repository"
)
__url__ = f"https://github.com/{__org__}/{__packagename__}"

DOWNLOAD_URL = f"https://github.com/{__org__}/{__packagename__}/archive/{__version__}.tar.gz"



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# STATEMENT OF CHANGES: This file is derived from sources licensed under the Apache-2.0 terms,
# and this file has been changed.
# The original file this work derives from is found at:
# https://github.com/nipreps/fmriprep/blob/a4fd718/fmriprep/workflows/bold/base.py
#
# [November 2021] CHANGES:
#   * Began tracking changes by copying over a4fd7a8 (21.0.0rc2) and noting differences
#   * Bold references are created with init_epi_reference_wf instead of init_bold_reference_wf.
#     Correspondingly, masking is performed separately. The initial boldref is performed in
#     workflows/base.py
#   * t1w is broadly replaced with anat
#   * Nibabies-specific subcortical alignment
#   * Carpet plots require CIFTI
#
# Copyright 2021 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""
Orchestrating the BOLD-preprocessing workflow
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

.. autofunction:: init_func_preproc_wf
.. autofunction:: init_func_derivatives_wf

"""
import os

import nibabel as nb
from nipype.interfaces import utility as niu
from nipype.interfaces.fsl import Split as FSLSplit
from nipype.pipeline import engine as pe
from niworkflows.interfaces.header import ValidateImage
from niworkflows.utils.connections import listify, pop_file
from sdcflows.interfaces.brainmask import BrainExtraction

from ... import config
from ...interfaces import DerivativesDataSink
from ...interfaces.reports import FunctionalSummary
from ...utils.bids import extract_entities
from ...utils.misc import combine_meepi_source

# BOLD workflows
from .confounds import init_bold_confs_wf, init_carpetplot_wf
from .hmc import init_bold_hmc_wf
from .outputs import init_func_derivatives_wf
from .registration import init_bold_reg_wf, init_bold_t1_trans_wf
from .resampling import (
    init_bold_preproc_trans_wf,
    init_bold_std_trans_wf,
    init_bold_surf_wf,
)
from .stc import init_bold_stc_wf
from .t2s import init_bold_t2s_wf


def init_func_preproc_wf(bold_file, has_fieldmap=False, existing_derivatives=None):
    """
    This workflow controls the functional preprocessing stages of *NiBabies*.

    Workflow Graph
        .. workflow::
            :graph2use: orig
            :simple_form: yes

            from fmriprep.workflows.tests import mock_config
            from nibabies import config
            from nibabies.workflows.bold.base import init_func_preproc_wf
            with mock_config():
                bold_file = config.execution.bids_dir / "sub-01" / "func" \
                    / "sub-01_task-mixedgamblestask_run-01_bold.nii.gz"
                wf = init_func_preproc_wf(str(bold_file))

    Parameters
    ----------
    bold_file
        Path to NIfTI file (single echo) or list of paths to NIfTI files (multi-echo)
    has_fieldmap : :obj:`bool`
        Signals the workflow to use inputnode fieldmap files

    Inputs
    ------
    bold_file
        BOLD series NIfTI file
    anat_preproc
        Bias-corrected structural template image
    anat_mask
        Mask of the skull-stripped template image
    anat_dseg
        Segmentation of preprocessed structural image, including
        gray-matter (GM), white-matter (WM) and cerebrospinal fluid (CSF)
    anat_aseg
        Segmentation of structural image, done with FreeSurfer.
    anat_aparc
        Parcellation of structural image, done with FreeSurfer.
    anat_tpms
        List of tissue probability maps in T1w space
    template
        List of templates to target
    anat2std_xfm
        List of transform files, collated with templates
    std2anat_xfm
        List of inverse transform files, collated with templates
    subjects_dir
        FreeSurfer SUBJECTS_DIR
    subject_id
        FreeSurfer subject ID
    t1w2fsnative_xfm
        LTA-style affine matrix translating from T1w to FreeSurfer-conformed subject space
    fsnative2t1w_xfm
        LTA-style affine matrix translating from FreeSurfer-conformed subject space to T1w
    bold_ref
        BOLD reference file
    bold_ref_xfm
        Transform file in LTA format from bold to reference
    n_dummy_scans
        Number of nonsteady states at the beginning of the BOLD run

    Outputs
    -------
    bold_t1
        BOLD series, resampled to T1w space
    bold_mask_t1
        BOLD series mask in T1w space
    bold_std
        BOLD series, resampled to template space
    bold_mask_std
        BOLD series mask in template space
    confounds
        TSV of confounds
    surfaces
        BOLD series, resampled to FreeSurfer surfaces
    aroma_noise_ics
        Noise components identified by ICA-AROMA
    melodic_mix
        FSL MELODIC mixing matrix
    bold_cifti
        BOLD CIFTI image
    cifti_variant
        combination of target spaces for `bold_cifti`

    See Also
    --------

    * :py:func:`~niworkflows.func.util.init_bold_reference_wf`
    * :py:func:`~nibabies.workflows.bold.stc.init_bold_stc_wf`
    * :py:func:`~nibabies.workflows.bold.hmc.init_bold_hmc_wf`
    * :py:func:`~nibabies.workflows.bold.t2s.init_bold_t2s_wf`
    * :py:func:`~nibabies.workflows.bold.registration.init_bold_t1_trans_wf`
    * :py:func:`~nibabies.workflows.bold.registration.init_bold_reg_wf`
    * :py:func:`~nibabies.workflows.bold.confounds.init_bold_confs_wf`
    * :py:func:`~nibabies.workflows.bold.confounds.init_ica_aroma_wf`
    * :py:func:`~nibabies.workflows.bold.resampling.init_bold_std_trans_wf`
    * :py:func:`~nibabies.workflows.bold.resampling.init_bold_preproc_trans_wf`
    * :py:func:`~nibabies.workflows.bold.resampling.init_bold_surf_wf`
    * :py:func:`~sdcflows.workflows.fmap.init_fmap_wf`
    * :py:func:`~sdcflows.workflows.pepolar.init_pepolar_unwarp_wf`
    * :py:func:`~sdcflows.workflows.phdiff.init_phdiff_wf`
    * :py:func:`~sdcflows.workflows.syn.init_syn_sdc_wf`
    * :py:func:`~sdcflows.workflows.unwarp.init_sdc_unwarp_wf`

    """
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow
    from niworkflows.interfaces.nibabel import ApplyMask
    from niworkflows.interfaces.utility import DictMerge, KeySelect
    from niworkflows.workflows.epi.refmap import init_epi_reference_wf

    if nb.load(bold_file[0] if isinstance(bold_file, (list, tuple)) else bold_file).shape[3:] <= (
        5 - config.execution.sloppy,
    ):
        config.loggers.workflow.warning(
            f"Too short BOLD series (<= 5 timepoints). Skipping processing of <{bold_file}>."
        )
        return

    mem_gb = {"filesize": 1, "resampled": 1, "largemem": 1}
    bold_tlen = 10

    # Have some options handy
    omp_nthreads = config.nipype.omp_nthreads
    freesurfer = config.workflow.run_reconall
    spaces = config.workflow.spaces
    nibabies_dir = str(config.execution.nibabies_dir)

    # Extract BIDS entities and metadata from BOLD file(s)
    entities = extract_entities(bold_file)
    layout = config.execution.layout

    # Extract metadata
    all_metadata = [layout.get_metadata(fname) for fname in listify(bold_file)]

    # Take first file as reference
    ref_file = pop_file(bold_file)
    metadata = all_metadata[0]
    # get original image orientation
    ref_orientation = get_img_orientation(ref_file)

    echo_idxs = listify(entities.get("echo", []))
    multiecho = len(echo_idxs) > 2
    if len(echo_idxs) == 1:
        config.loggers.workflow.warning(
            f"Running a single echo <{ref_file}> from a seemingly multi-echo dataset."
        )
        bold_file = ref_file  # Just in case - drop the list

    if len(echo_idxs) == 2:
        raise RuntimeError(
            "Multi-echo processing requires at least three different echos (found two)."
        )

    if multiecho:
        # Drop echo entity for future queries, have a boolean shorthand
        entities.pop("echo", None)
        # reorder echoes from shortest to largest
        tes, bold_file = zip(
            *sorted([(layout.get_metadata(bf, scope="raw")["EchoTime"], bf) for bf in bold_file])
        )
        ref_file = bold_file[0]  # Reset reference to be the shortest TE

    if os.path.isfile(ref_file):
        bold_tlen, mem_gb = _create_mem_gb(ref_file)

    wf_name = _get_wf_name(ref_file)
    config.loggers.workflow.debug(
        f'Creating bold processing workflow for <{ref_file}> ({mem_gb["filesize"]:.2f} GB '
        f'/ {bold_tlen} TRs). Memory resampled/largemem={mem_gb["resampled"]:.2f}'
        f'/{mem_gb["largemem"]:.2f} GB.'
    )

    # Find associated sbref, if possible
    entities["suffix"] = "sbref"
    entities["extension"] = [".nii", ".nii.gz"]  # Overwrite extensions
    sbref_files = layout.get(scope="raw", return_type="file", **entities)

    sbref_msg = f"No single-band-reference found for {os.path.basename(ref_file)}."
    if sbref_files and "sbref" in config.workflow.ignore:
        sbref_msg = "Single-band reference file(s) found and ignored."
        sbref_files = []
    elif sbref_files:
        sbref_msg = "Using single-band reference file(s) {}.".format(
            ",".join([os.path.basename(sbf) for sbf in sbref_files])
        )
    config.loggers.workflow.info(sbref_msg)

    if has_fieldmap:
        # First check if specified via B0FieldSource
        estimator_key = listify(metadata.get("B0FieldSource"))

        if not estimator_key:
            import re
            from pathlib import Path

            from sdcflows.fieldmaps import get_identifier

            # Fallback to IntendedFor
            intended_rel = re.sub(
                r"^sub-[a-zA-Z0-9]*/",
                "",
                str(Path(bold_file if not multiecho else bold_file[0]).relative_to(layout.root)),
            )
            estimator_key = get_identifier(intended_rel)

        if not estimator_key:
            has_fieldmap = False
            config.loggers.workflow.critical(
                f"None of the available B0 fieldmaps are associated to <{bold_file}>"
            )
        else:
            config.loggers.workflow.info(
                f"Found usable B0-map (fieldmap) estimator(s) <{', '.join(estimator_key)}> "
                f"to correct <{bold_file}> for susceptibility-derived distortions."
            )

    # Check whether STC must/can be run
    run_stc = bool(metadata.get("SliceTiming")) and "slicetiming" not in config.workflow.ignore

    # Build workflow
    workflow = Workflow(name=wf_name)
    workflow.__postdesc__ = """\
All resamplings can be performed with *a single interpolation
step* by composing all the pertinent transformations (i.e. head-motion
transform matrices, susceptibility distortion correction when available,
and co-registrations to anatomical and output spaces).
Gridded (volumetric) resamplings were performed using `antsApplyTransforms` (ANTs),
configured with Lanczos interpolation to minimize the smoothing
effects of other kernels [@lanczos].
Non-gridded (surface) resamplings were performed using `mri_vol2surf`
(FreeSurfer).
"""

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                "bold_file",
                # from smriprep
                "anat_preproc",
                "anat_brain",
                "anat_mask",
                "anat_dseg",
                "anat_tpms",
                "anat_aseg",
                "anat_aparc",
                "anat2std_xfm",
                "std2anat_xfm",
                "template",
                # from bold reference workflow
                "bold_ref",
                "bold_ref_xfm",
                "n_dummy_scans",
                # from sdcflows (optional)
                "fmap",
                "fmap_ref",
                "fmap_coeff",
                "fmap_mask",
                "fmap_id",
                "sdc_method",
                # if reconstructing with FreeSurfer (optional)
                "t1w2fsnative_xfm",
                "fsnative2t1w_xfm",
                "subject_id",
                "subjects_dir",
            ]
        ),
        name="inputnode",
    )
    inputnode.inputs.bold_file = bold_file

    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                "bold_anat",
                "bold_anat_ref",
                "bold2anat_xfm",
                "anat2bold_xfm",
                "bold_mask_anat",
                "bold_aseg_anat",
                "bold_aparc_anat",
                "bold_std",
                "bold_std_ref",
                "bold_mask_std",
                "bold_aseg_std",
                "bold_aparc_std",
                "bold_native",
                "bold_native_ref",
                "bold_mask_native",
                "bold_echos_native",
                "bold_cifti",
                "cifti_variant",
                "cifti_metadata",
                "cifti_density",
                "surfaces",
                "confounds",
                "aroma_noise_ics",
                "melodic_mix",
                "nonaggr_denoised_file",
                "confounds_metadata",
            ]
        ),
        name="outputnode",
    )

    # Generate a brain-masked conversion of the t1w
    t1w_brain = pe.Node(ApplyMask(), name="t1w_brain")

    # Track echo index - this allows us to treat multi- and single-echo workflows
    # almost identically
    echo_index = pe.Node(niu.IdentityInterface(fields=["echoidx"]), name="echo_index")
    if multiecho:
        echo_index.iterables = [("echoidx", range(len(bold_file)))]
    else:
        echo_index.inputs.echoidx = 0

    # BOLD source: track original BOLD file(s)
    bold_source = pe.Node(
        niu.Select(inlist=bold_file), run_without_submitting=True, name="bold_source"
    )

    # BOLD buffer: an identity used as a pointer to either the original BOLD
    # or the STC'ed one for further use.
    boldbuffer = pe.Node(niu.IdentityInterface(fields=["bold_file"]), name="boldbuffer")

    summary = pe.Node(
        FunctionalSummary(
            slice_timing=run_stc,
            registration=("FSL", "FreeSurfer")[freesurfer],
            registration_dof=config.workflow.bold2t1w_dof,
            registration_init=config.workflow.bold2t1w_init,
            pe_direction=metadata.get("PhaseEncodingDirection"),
            echo_idx=echo_idxs,
            tr=metadata["RepetitionTime"],
            orientation=ref_orientation,
        ),
        name="summary",
        mem_gb=config.DEFAULT_MEMORY_MIN_GB,
        run_without_submitting=True,
    )
    summary.inputs.dummy_scans = config.workflow.dummy_scans

    func_derivatives_wf = init_func_derivatives_wf(
        bids_root=layout.root,
        cifti_output=config.workflow.cifti_output,
        freesurfer=freesurfer,
        all_metadata=all_metadata,
        multiecho=multiecho,
        output_dir=nibabies_dir,
        spaces=spaces,
        use_aroma=config.workflow.use_aroma,
    )
    func_derivatives_wf.inputs.inputnode.all_source_files = bold_file

    # fmt:off
    workflow.connect([
        (outputnode, func_derivatives_wf, [
            ('bold_anat', 'inputnode.bold_t1'),
            ('bold_anat_ref', 'inputnode.bold_t1_ref'),
            ('bold2anat_xfm', 'inputnode.bold2anat_xfm'),
            ('anat2bold_xfm', 'inputnode.anat2bold_xfm'),
            ('bold_aseg_anat', 'inputnode.bold_aseg_t1'),
            ('bold_aparc_anat', 'inputnode.bold_aparc_t1'),
            ('bold_mask_anat', 'inputnode.bold_mask_t1'),
            ('bold_native', 'inputnode.bold_native'),
            ("bold_native_ref", "inputnode.bold_native_ref"),
            ("bold_mask_native", "inputnode.bold_mask_native"),
            ("bold_echos_native", "inputnode.bold_echos_native"),
            ('confounds', 'inputnode.confounds'),
            ('surfaces', 'inputnode.surf_files'),
            ('aroma_noise_ics', 'inputnode.aroma_noise_ics'),
            ('melodic_mix', 'inputnode.melodic_mix'),
            ('nonaggr_denoised_file', 'inputnode.nonaggr_denoised_file'),
            ('bold_cifti', 'inputnode.bold_cifti'),
            ('cifti_variant', 'inputnode.cifti_variant'),
            ('cifti_metadata', 'inputnode.cifti_metadata'),
            ('cifti_density', 'inputnode.cifti_density'),
            ('confounds_metadata', 'inputnode.confounds_metadata'),
            ('acompcor_masks', 'inputnode.acompcor_masks'),
            ('tcompcor_mask', 'inputnode.tcompcor_mask'),
        ]),
    ])
    # fmt:on

    # Extract BOLD validation from init_bold_reference_wf
    validate_bolds = pe.MapNode(
        ValidateImage(),
        name="validate_bolds",
        mem_gb=config.DEFAULT_MEMORY_MIN_GB,
        iterfield=["in_file"],
    )
    validate_bolds.inputs.in_file = listify(bold_file)

    # Select validated BOLD files (orientations checked or corrected)
    select_bold = pe.Node(niu.Select(), run_without_submitting=True, name="select_bold")

    # Top-level BOLD splitter
    bold_split = pe.Node(FSLSplit(dimension="t"), name="bold_split", mem_gb=mem_gb["filesize"] * 3)

    # HMC on the BOLD
    bold_hmc_wf = init_bold_hmc_wf(
        name="bold_hmc_wf", mem_gb=mem_gb["filesize"], omp_nthreads=omp_nthreads
    )

    # calculate BOLD registration to T1w
    bold_reg_wf = init_bold_reg_wf(
        bold2t1w_dof=config.workflow.bold2t1w_dof,
        bold2t1w_init=config.workflow.bold2t1w_init,
        freesurfer=freesurfer,
        mem_gb=mem_gb["resampled"],
        name="bold_reg_wf",
        omp_nthreads=omp_nthreads,
        sloppy=config.execution.sloppy,
        use_bbr=config.workflow.use_bbr,
    )

    # apply BOLD registration to T1w
    bold_t1_trans_wf = init_bold_t1_trans_wf(
        name="bold_t1_trans_wf",
        freesurfer=freesurfer,
        mem_gb=mem_gb["resampled"],
        omp_nthreads=omp_nthreads,
        use_compression=False,
    )
    bold_t1_trans_wf.inputs.inputnode.fieldwarp = "identity"

    # get confounds
    bold_confounds_wf = init_bold_confs_wf(
        mem_gb=mem_gb["largemem"],
        metadata=metadata,
        freesurfer=freesurfer,
        regressors_all_comps=config.workflow.regressors_all_comps,
        regressors_fd_th=config.workflow.regressors_fd_th,
        regressors_dvars_th=config.workflow.regressors_dvars_th,
        fd_radius=config.workflow.fd_radius,
        name="bold_confounds_wf",
    )
    bold_confounds_wf.get_node("inputnode").inputs.t1_transform_flags = [False]

    # SLICE-TIME CORRECTION (or bypass) #############################################
    if run_stc:
        bold_stc_wf = init_bold_stc_wf(name="bold_stc_wf", metadata=metadata)
        # fmt:off
        workflow.connect([
            (inputnode, bold_stc_wf, [('n_dummy_scans', 'inputnode.skip_vols')]),
            (select_bold, bold_stc_wf, [("out", 'inputnode.bold_file')]),
            (bold_stc_wf, boldbuffer, [('outputnode.stc_file', 'bold_file')]),
        ])
        # fmt:on

    # bypass STC from original BOLD in both SE and ME cases
    else:
        workflow.connect([(select_bold, boldbuffer, [("out", "bold_file")])])

    # MULTI-ECHO EPI DATA #############################################
    if multiecho:  # instantiate relevant interfaces, imports
        split_opt_comb = bold_split.clone(name="split_opt_comb")

        inputnode.inputs.bold_file = ref_file  # Replace reference w first echo

        join_echos = pe.JoinNode(
            niu.IdentityInterface(fields=["bold_files"]),
            joinsource="echo_index",
            joinfield=["bold_files"],
            name="join_echos",
        )

        # create optimal combination, adaptive T2* map
        bold_t2s_wf = init_bold_t2s_wf(
            echo_times=tes,
            mem_gb=mem_gb["resampled"],
            omp_nthreads=omp_nthreads,
            name="bold_t2smap_wf",
        )

    bold_final = pe.Node(
        niu.IdentityInterface(fields=["bold", "boldref", "mask", "bold_echos"]), name="bold_final"
    )

    # Mask input BOLD reference image
    initial_boldref_mask = pe.Node(BrainExtraction(), name="initial_boldref_mask")

    # This final boldref will be calculated after bold_bold_trans_wf, which includes one or more:
    # HMC (head motion correction)
    # STC (slice time correction)
    # SDC (susceptibility distortion correction)
    final_boldref_wf = init_epi_reference_wf(
        auto_bold_nss=True,
        omp_nthreads=omp_nthreads,
        name="final_boldref_wf",
    )
    final_boldref_mask = pe.Node(BrainExtraction(), name="final_boldref_mask")

    # MAIN WORKFLOW STRUCTURE #######################################################
    # fmt:off
    workflow.connect([
        # Prepare masked T1w image
        (inputnode, t1w_brain, [("anat_preproc", "in_file"),
                                ("anat_mask", "in_mask")]),
        # Select validated bold files per-echo
        (validate_bolds, select_bold, [("out_file", "inlist")]),
        # BOLD buffer has slice-time corrected if it was run, original otherwise
        (boldbuffer, bold_split, [('bold_file', 'in_file')]),
        # HMC
        (inputnode, bold_hmc_wf, [
            ('bold_ref', 'inputnode.raw_ref_image')]),
        (validate_bolds, bold_hmc_wf, [
            (("out_file", pop_file), 'inputnode.bold_file')]),
        # Native-space BOLD files
        (final_boldref_wf, final_boldref_mask, [('outputnode.epi_ref_file', 'in_file')]),
        (final_boldref_wf, bold_final, [('outputnode.epi_ref_file', 'boldref')]),
        (final_boldref_mask, bold_final, [('out_file', 'mask')]),
        (bold_final, outputnode, [
            ('bold', 'bold_native'),
            ('boldref', 'bold_native_ref'),
            ('mask', 'bold_mask_native'),
            ('bold_echos', 'bold_echos_native'),
        ]),
        # EPI-T1 registration workflow
        (inputnode, bold_reg_wf, [
            ('anat_dseg', 'inputnode.t1w_dseg'),
            # Undefined if --fs-no-reconall, but this is safe
            ('subjects_dir', 'inputnode.subjects_dir'),
            ('subject_id', 'inputnode.subject_id'),
            ('fsnative2t1w_xfm', 'inputnode.fsnative2t1w_xfm'),
        ]),
        (bold_final, bold_reg_wf, [
            ("boldref", "inputnode.ref_bold_brain")]),
        (t1w_brain, bold_reg_wf, [("out_file", "inputnode.t1w_brain")]),
        (inputnode, bold_t1_trans_wf, [
            ('bold_file', 'inputnode.name_source'),
            ('anat_mask', 'inputnode.t1w_mask'),
            ('anat_aseg', 'inputnode.t1w_aseg'),
            ("anat_aparc", "inputnode.t1w_aparc"),
        ]),
        (t1w_brain, bold_t1_trans_wf, [("out_file", "inputnode.t1w_brain")]),
        (bold_reg_wf, outputnode, [
            ('outputnode.itk_bold_to_t1', 'bold2anat_xfm'),
            ('outputnode.itk_t1_to_bold', 'anat2bold_xfm')]),
        (bold_reg_wf, bold_t1_trans_wf, [
            ('outputnode.itk_bold_to_t1', 'inputnode.itk_bold_to_t1')]),
        (bold_final, bold_t1_trans_wf, [
            ('mask', 'inputnode.ref_bold_mask'),
            ('boldref', 'inputnode.ref_bold_brain'),
        ]),
        (bold_t1_trans_wf, outputnode, [('outputnode.bold_t1', 'bold_anat'),
                                        ('outputnode.bold_t1_ref', 'bold_anat_ref'),
                                        ('outputnode.bold_aseg_t1', 'bold_aseg_anat'),
                                        ('outputnode.bold_aparc_t1', 'bold_aparc_anat')]),
        # Connect bold_confounds_wf
        (inputnode, bold_confounds_wf, [('anat_tpms', 'inputnode.t1w_tpms'),
                                        ('anat_mask', 'inputnode.t1w_mask')]),
        (bold_hmc_wf, bold_confounds_wf, [
            ('outputnode.movpar_file', 'inputnode.movpar_file'),
            ('outputnode.rmsd_file', 'inputnode.rmsd_file')]),
        (bold_reg_wf, bold_confounds_wf, [
            ('outputnode.itk_t1_to_bold', 'inputnode.t1_bold_xform')]),
        (inputnode, bold_confounds_wf, [
            ('n_dummy_scans', 'inputnode.skip_vols')]),
        (bold_final, bold_confounds_wf, [
            ('bold', 'inputnode.bold'),
            ('mask', 'inputnode.bold_mask'),
        ]),
        (bold_confounds_wf, outputnode, [
            ('outputnode.confounds_file', 'confounds'),
            ('outputnode.confounds_metadata', 'confounds_metadata'),
            ('outputnode.acompcor_masks', 'acompcor_masks'),
            ('outputnode.tcompcor_mask', 'tcompcor_mask'),
        ]),
        # Summary
        (inputnode, summary, [('n_dummy_scans', 'algo_dummy_scans')]),
        (bold_reg_wf, summary, [('outputnode.fallback', 'fallback')]),
        (outputnode, summary, [('confounds', 'confounds_file')]),
        # Select echo indices for original/validated BOLD files
        (echo_index, bold_source, [("echoidx", "index")]),
        (echo_index, select_bold, [("echoidx", "index")]),
    ])
    # fmt:on

    # for standard EPI data, pass along correct file
    if not multiecho:
        # fmt:off
        workflow.connect([
            (inputnode, func_derivatives_wf, [
                ('bold_file', 'inputnode.source_file')]),
            (bold_split, bold_t1_trans_wf, [
                ('out_files', 'inputnode.bold_split')]),
            (bold_hmc_wf, bold_t1_trans_wf, [
                ('outputnode.xforms', 'inputnode.hmc_xforms')]),
        ])
        # fmt:on
    else:  # for meepi, use optimal combination
        # fmt:off
        workflow.connect([
            # update name source for optimal combination
            (inputnode, func_derivatives_wf, [
                (('bold_file', combine_meepi_source), 'inputnode.source_file')]),
            (join_echos, bold_t2s_wf, [
                ('bold_files', 'inputnode.bold_file')]),
            (bold_t2s_wf, split_opt_comb, [
                ('outputnode.bold', 'in_file')]),
            (split_opt_comb, bold_t1_trans_wf, [
                ("out_files", "inputnode.bold_split")]),
            (join_echos, bold_final, [("bold_files", "bold_echos")]),
            (bold_t2s_wf, bold_final, [("outputnode.bold", "bold")]),
        ])
        # fmt:on

        # Already applied in bold_bold_trans_wf, which inputs to bold_t2s_wf
        bold_t1_trans_wf.inputs.inputnode.hmc_xforms = "identity"

    # Map final BOLD mask into T1w space (if required)
    nonstd_spaces = set(spaces.get_nonstandard())
    if nonstd_spaces.intersection(("T1w", "anat")):
        from niworkflows.interfaces.fixes import (
            FixHeaderApplyTransforms as ApplyTransforms,
        )

        boldmask_to_t1w = pe.Node(
            ApplyTransforms(interpolation="MultiLabel"), name="boldmask_to_t1w", mem_gb=0.1
        )
        # fmt:off
        workflow.connect([
            (bold_reg_wf, boldmask_to_t1w, [
                ('outputnode.itk_bold_to_t1', 'transforms')]),
            (bold_t1_trans_wf, boldmask_to_t1w, [
                ('outputnode.bold_mask_t1', 'reference_image')]),
            (boldmask_to_t1w, outputnode, [
                ('output_image', 'bold_mask_anat')]),
        ])
        # fmt:on

    if spaces.get_spaces(nonstandard=False, dim=(3,)):
        # Apply transforms in 1 shot
        # Only use uncompressed output if AROMA is to be run
        bold_std_trans_wf = init_bold_std_trans_wf(
            freesurfer=freesurfer,
            mem_gb=mem_gb["resampled"],
            omp_nthreads=omp_nthreads,
            spaces=spaces,
            name="bold_std_trans_wf",
            use_compression=not config.execution.low_mem,
        )
        bold_std_trans_wf.inputs.inputnode.fieldwarp = "identity"

        # fmt:off
        workflow.connect([
            (inputnode, bold_std_trans_wf, [
                ('template', 'inputnode.templates'),
                ('anat2std_xfm', 'inputnode.anat2std_xfm'),
                ('bold_file', 'inputnode.name_source'),
                ('anat_aseg', 'inputnode.bold_aseg'),
                ('anat_aparc', 'inputnode.bold_aparc')]),
            (bold_final, bold_std_trans_wf, [
                ('mask', 'inputnode.bold_mask'),
            ]),
            (bold_reg_wf, bold_std_trans_wf, [
                ('outputnode.itk_bold_to_t1', 'inputnode.itk_bold_to_t1')]),
            (bold_std_trans_wf, outputnode, [('outputnode.bold_std', 'bold_std'),
                                             ('outputnode.bold_std_ref', 'bold_std_ref'),
                                             ('outputnode.bold_mask_std', 'bold_mask_std')]),
        ])
        # fmt:on

        if freesurfer:
            # fmt:off
            workflow.connect([
                (bold_std_trans_wf, func_derivatives_wf, [
                    ('outputnode.bold_aseg_std', 'inputnode.bold_aseg_std'),
                    ('outputnode.bold_aparc_std', 'inputnode.bold_aparc_std'),
                ]),
                (bold_std_trans_wf, outputnode, [
                    ('outputnode.bold_aseg_std', 'bold_aseg_std'),
                    ('outputnode.bold_aparc_std', 'bold_aparc_std')]),
            ])
            # fmt:on

        if not multiecho:
            # fmt:off
            workflow.connect([
                (bold_split, bold_std_trans_wf, [
                    ('out_files', 'inputnode.bold_split')]),
                (bold_hmc_wf, bold_std_trans_wf, [
                    ('outputnode.xforms', 'inputnode.hmc_xforms')]),
            ])
            # fmt:on
        else:
            # fmt:off
            workflow.connect([
                (split_opt_comb, bold_std_trans_wf, [
                    ('out_files', 'inputnode.bold_split')])
            ])
            # fmt:on

            # Already applied in bold_bold_trans_wf, which inputs to bold_t2s_wf
            bold_std_trans_wf.inputs.inputnode.hmc_xforms = "identity"

        # fmt:off
        # func_derivatives_wf internally parametrizes over snapshotted spaces.
        workflow.connect([
            (bold_std_trans_wf, func_derivatives_wf, [
                ('outputnode.template', 'inputnode.template'),
                ('outputnode.spatial_reference', 'inputnode.spatial_reference'),
                ('outputnode.bold_std_ref', 'inputnode.bold_std_ref'),
                ('outputnode.bold_std', 'inputnode.bold_std'),
                ('outputnode.bold_mask_std', 'inputnode.bold_mask_std'),
            ]),
        ])
        # fmt:on

        if config.workflow.use_aroma:  # ICA-AROMA workflow
            from .confounds import init_ica_aroma_wf

            ica_aroma_wf = init_ica_aroma_wf(
                mem_gb=mem_gb["resampled"],
                metadata=metadata,
                omp_nthreads=omp_nthreads,
                err_on_aroma_warn=config.workflow.aroma_err_on_warn,
                aroma_melodic_dim=config.workflow.aroma_melodic_dim,
                name="ica_aroma_wf",
            )

            join = pe.Node(
                niu.Function(output_names=["out_file"], function=_to_join), name="aroma_confounds"
            )

            mrg_conf_metadata = pe.Node(
                niu.Merge(2), name="merge_confound_metadata", run_without_submitting=True
            )
            mrg_conf_metadata2 = pe.Node(
                DictMerge(), name="merge_confound_metadata2", run_without_submitting=True
            )
            # fmt:off
            workflow.disconnect([
                (bold_confounds_wf, outputnode, [
                    ('outputnode.confounds_file', 'confounds'),
                ]),
                (bold_confounds_wf, outputnode, [
                    ('outputnode.confounds_metadata', 'confounds_metadata'),
                ]),
            ])
            workflow.connect([
                (inputnode, ica_aroma_wf, [
                    ('bold_file', 'inputnode.name_source')]),
                (bold_hmc_wf, ica_aroma_wf, [
                    ('outputnode.movpar_file', 'inputnode.movpar_file')]),
                (inputnode, ica_aroma_wf, [
                    ('n_dummy_scans', 'inputnode.skip_vols')]),
                (bold_confounds_wf, join, [
                    ('outputnode.confounds_file', 'in_file')]),
                (bold_confounds_wf, mrg_conf_metadata,
                    [('outputnode.confounds_metadata', 'in1')]),
                (ica_aroma_wf, join,
                    [('outputnode.aroma_confounds', 'join_file')]),
                (ica_aroma_wf, mrg_conf_metadata,
                    [('outputnode.aroma_metadata', 'in2')]),
                (mrg_conf_metadata, mrg_conf_metadata2, [('out', 'in_dicts')]),
                (ica_aroma_wf, outputnode,
                    [('outputnode.aroma_noise_ics', 'aroma_noise_ics'),
                     ('outputnode.melodic_mix', 'melodic_mix'),
                     ('outputnode.nonaggr_denoised_file', 'nonaggr_denoised_file')]),
                (join, outputnode, [('out_file', 'confounds')]),
                (mrg_conf_metadata2, outputnode, [('out_dict', 'confounds_metadata')]),
                (bold_std_trans_wf, ica_aroma_wf, [
                    ('outputnode.bold_std', 'inputnode.bold_std'),
                    ('outputnode.bold_mask_std', 'inputnode.bold_mask_std'),
                    ('outputnode.spatial_reference', 'inputnode.spatial_reference')]),
            ])
            # fmt:on

    # SURFACES ##################################################################################
    # Freesurfer
    freesurfer_spaces = spaces.get_fs_spaces()
    if freesurfer and freesurfer_spaces:
        config.loggers.workflow.debug("Creating BOLD surface-sampling workflow.")
        bold_surf_wf = init_bold_surf_wf(
            mem_gb=mem_gb["resampled"],
            surface_spaces=freesurfer_spaces,
            medial_surface_nan=config.workflow.medial_surface_nan,
            name="bold_surf_wf",
        )
        # fmt:off
        workflow.connect([
            (inputnode, bold_surf_wf, [
                ('subjects_dir', 'inputnode.subjects_dir'),
                ('subject_id', 'inputnode.subject_id'),
                ('t1w2fsnative_xfm', 'inputnode.t1w2fsnative_xfm')]),
            (bold_t1_trans_wf, bold_surf_wf, [('outputnode.bold_t1', 'inputnode.source_file')]),
            (bold_surf_wf, outputnode, [('outputnode.surfaces', 'surfaces')]),
            (bold_surf_wf, func_derivatives_wf, [
                ('outputnode.target', 'inputnode.surf_refs')]),
        ])
        # fmt:on

        # CIFTI output
        if config.workflow.cifti_output:
            from .alignment import (
                init_subcortical_mni_alignment_wf,
                init_subcortical_rois_wf,
            )
            from .resampling import init_bold_grayords_wf

            key = get_MNIInfant_key(spaces)

            # BOLD/ROIs should be in MNIInfant space
            cifti_select_std = pe.Node(
                KeySelect(fields=["bold_std", "bold_aseg_std"], key=key),
                name="cifti_select_std",
                run_without_submitting=True,
            )

            subcortical_rois_wf = init_subcortical_rois_wf()
            subcortical_mni_alignment_wf = init_subcortical_mni_alignment_wf()
            bold_grayords_wf = init_bold_grayords_wf(
                grayord_density=config.workflow.cifti_output,
                mem_gb=mem_gb["resampled"],
                repetition_time=metadata["RepetitionTime"],
            )

            # fmt:off
            workflow.connect([
                (bold_std_trans_wf, cifti_select_std, [
                    ("outputnode.bold_std", "bold_std"),
                    ("outputnode.bold_aseg_std", "bold_aseg_std"),
                    ("outputnode.spatial_reference", "keys")]),
                (cifti_select_std, subcortical_rois_wf, [
                    ("bold_aseg_std", "inputnode.MNIInfant_aseg")]),
                (cifti_select_std, subcortical_mni_alignment_wf, [
                    ("bold_std", "inputnode.MNIInfant_bold")]),
                (subcortical_rois_wf, subcortical_mni_alignment_wf, [
                    ("outputnode.MNIInfant_rois", "inputnode.MNIInfant_rois"),
                    ("outputnode.MNI152_rois", "inputnode.MNI152_rois")]),
                (subcortical_mni_alignment_wf, bold_grayords_wf, [
                    ("outputnode.subcortical_volume", "inputnode.subcortical_volume"),
                    ("outputnode.subcortical_labels", "inputnode.subcortical_labels")]),
                (bold_surf_wf, bold_grayords_wf, [
                    ('outputnode.surfaces', 'inputnode.surf_files'),
                    ('outputnode.target', 'inputnode.surf_refs')]),
                (bold_grayords_wf, outputnode, [
                    ('outputnode.cifti_bold', 'bold_cifti'),
                    ('outputnode.cifti_variant', 'cifti_variant'),
                    ('outputnode.cifti_metadata', 'cifti_metadata'),
                    ('outputnode.cifti_density', 'cifti_density')]),
            ])
            # fmt:on

    if spaces.get_spaces(nonstandard=False, dim=(3,)):
        if not config.workflow.cifti_output:
            config.loggers.workflow.critical("The carpetplot requires CIFTI outputs")
        else:
            carpetplot_wf = init_carpetplot_wf(
                mem_gb=mem_gb["resampled"],
                metadata=metadata,
                cifti_output=bool(config.workflow.cifti_output),
                name="carpetplot_wf",
            )

            # fmt:off
            workflow.connect([
                (bold_grayords_wf, carpetplot_wf, [
                    ('outputnode.cifti_bold', 'inputnode.cifti_bold')]),
                (bold_confounds_wf, carpetplot_wf, [
                    ('outputnode.confounds_file', 'inputnode.confounds_file')]),
            ])
            # fmt:on

    # REPORTING ############################################################
    ds_report_summary = pe.Node(
        DerivativesDataSink(desc="summary", datatype="figures", dismiss_entities=("echo",)),
        name="ds_report_summary",
        run_without_submitting=True,
        mem_gb=config.DEFAULT_MEMORY_MIN_GB,
    )

    ds_report_validation = pe.Node(
        DerivativesDataSink(desc="validation", datatype="figures", dismiss_entities=("echo",)),
        name="ds_report_validation",
        run_without_submitting=True,
        mem_gb=config.DEFAULT_MEMORY_MIN_GB,
    )

    # fmt:off
    workflow.connect([
        (summary, ds_report_summary, [('out_report', 'in_file')]),
        (validate_bolds, ds_report_validation, [
            (("out_report", pop_file), 'in_file')]),
    ])
    # fmt:on

    # Fill-in datasinks of reportlets seen so far
    for node in workflow.list_node_names():
        if node.split(".")[-1].startswith("ds_report"):
            workflow.get_node(node).inputs.base_directory = nibabies_dir
            workflow.get_node(node).inputs.source_file = ref_file

    if not has_fieldmap:
        # Finalize workflow without SDC connections
        summary.inputs.distortion_correction = "None"

        # Resample in native space in just one shot
        bold_bold_trans_wf = init_bold_preproc_trans_wf(
            mem_gb=mem_gb["resampled"],
            omp_nthreads=omp_nthreads,
            use_compression=not config.execution.low_mem,
            use_fieldwarp=False,
            name="bold_bold_trans_wf",
        )
        bold_bold_trans_wf.inputs.inputnode.fieldwarp = "identity"

        # fmt:off
        workflow.connect([
            # Connect bold_bold_trans_wf
            (bold_source, bold_bold_trans_wf, [("out", "inputnode.name_source")]),
            (bold_split, bold_bold_trans_wf, [("out_files", "inputnode.bold_file")]),
            (bold_hmc_wf, bold_bold_trans_wf, [
                ("outputnode.xforms", "inputnode.hmc_xforms"),
            ]),
        ])

        workflow.connect([
            (bold_bold_trans_wf, bold_final, [("outputnode.bold", "bold")]),
            (bold_bold_trans_wf, final_boldref_wf, [
                ("outputnode.bold", "inputnode.in_files"),
            ]),
        ] if not multiecho else [
            (inputnode, initial_boldref_mask, [('bold_ref', 'in_file')]),
            (initial_boldref_mask, bold_t2s_wf, [
                ("out_file", "inputnode.bold_mask"),
            ]),
            (bold_bold_trans_wf, join_echos, [
                ("outputnode.bold", "bold_files"),
            ]),
            (join_echos, final_boldref_wf, [
                ("bold_files", "inputnode.in_files"),
            ]),
        ])
        # fmt:on
        return workflow

    from niworkflows.interfaces.reportlets.registration import (
        SimpleBeforeAfterRPT as SimpleBeforeAfter,
    )
    from niworkflows.interfaces.utility import KeySelect
    from sdcflows.workflows.apply.correction import init_unwarp_wf
    from sdcflows.workflows.apply.registration import init_coeff2epi_wf

    coeff2epi_wf = init_coeff2epi_wf(
        debug="fieldmaps" in config.execution.debug,
        omp_nthreads=config.nipype.omp_nthreads,
        write_coeff=True,
    )
    unwarp_wf = init_unwarp_wf(
        debug="fieldmaps" in config.execution.debug, omp_nthreads=config.nipype.omp_nthreads
    )
    unwarp_wf.inputs.inputnode.metadata = metadata

    output_select = pe.Node(
        KeySelect(fields=["fmap", "fmap_ref", "fmap_coeff", "fmap_mask", "sdc_method"]),
        name="output_select",
        run_without_submitting=True,
    )
    output_select.inputs.key = estimator_key[0]
    if len(estimator_key) > 1:
        config.loggers.workflow.warning(
            f"Several fieldmaps <{', '.join(estimator_key)}> are "
            f"'IntendedFor' <{bold_file}>, using {estimator_key[0]}"
        )

    sdc_report = pe.Node(
        SimpleBeforeAfter(before_label="Distorted", after_label="Corrected", dismiss_affine=True),
        name="sdc_report",
        mem_gb=0.1,
    )

    ds_report_sdc = pe.Node(
        DerivativesDataSink(
            base_directory=nibabies_dir,
            desc="sdc",
            suffix="bold",
            datatype="figures",
            dismiss_entities=("echo",),
        ),
        name="ds_report_sdc",
        run_without_submitting=True,
    )

    # fmt:off
    workflow.connect([
        (inputnode, output_select, [("fmap", "fmap"),
                                    ("fmap_ref", "fmap_ref"),
                                    ("fmap_coeff", "fmap_coeff"),
                                    ("fmap_mask", "fmap_mask"),
                                    ("sdc_method", "sdc_method"),
                                    ("fmap_id", "keys")]),
        (output_select, coeff2epi_wf, [
            ("fmap_ref", "inputnode.fmap_ref"),
            ("fmap_coeff", "inputnode.fmap_coeff"),
            ("fmap_mask", "inputnode.fmap_mask")]),
        (output_select, summary, [("sdc_method", "distortion_correction")]),
        (inputnode, initial_boldref_mask, [('bold_ref', 'in_file')]),
        (inputnode, coeff2epi_wf, [
            ("bold_ref", "inputnode.target_ref")]),
        (initial_boldref_mask, coeff2epi_wf, [
            ("out_file", "inputnode.target_mask")]),  # skull-stripped brain
        (coeff2epi_wf, unwarp_wf, [
            ("outputnode.fmap_coeff", "inputnode.fmap_coeff")]),
        (inputnode, sdc_report, [("bold_ref", "before")]),
        (bold_hmc_wf, unwarp_wf, [
            ("outputnode.xforms", "inputnode.hmc_xforms")]),
        (bold_split, unwarp_wf, [
            ("out_files", "inputnode.distorted")]),
        (final_boldref_wf, sdc_report, [("outputnode.epi_ref_file", "after")]),
        (final_boldref_mask, sdc_report, [("out_file", "wm_seg")]),
        (inputnode, ds_report_sdc, [("bold_file", "source_file")]),
        (sdc_report, ds_report_sdc, [("out_report", "in_file")]),

    ])
    # fmt:on

    if not multiecho:
        # fmt:off
        workflow.connect([
            (unwarp_wf, bold_final, [("outputnode.corrected", "bold")]),
            # remaining workflow connections
            (unwarp_wf, final_boldref_wf, [
                ("outputnode.corrected", "inputnode.in_files"),
            ]),
            (unwarp_wf, bold_t1_trans_wf, [
                # TEMPORARY: For the moment we can't use frame-wise fieldmaps
                (("outputnode.fieldwarp", pop_file), "inputnode.fieldwarp"),
            ]),
            (unwarp_wf, bold_std_trans_wf, [
                # TEMPORARY: For the moment we can't use frame-wise fieldmaps
                (("outputnode.fieldwarp", pop_file), "inputnode.fieldwarp"),
            ]),
        ])
        # fmt:on
        return workflow

    # Finalize connections if ME-EPI
    join_sdc_echos = pe.JoinNode(
        niu.IdentityInterface(
            fields=[
                "fieldmap",
                "fieldwarp",
                "corrected",
                "corrected_ref",
                "corrected_mask",
            ]
        ),
        joinsource="echo_index",
        joinfield=[
            "fieldmap",
            "fieldwarp",
            "corrected",
            "corrected_ref",
            "corrected_mask",
        ],
        name="join_sdc_echos",
    )

    def _dpop(list_of_lists):
        return list_of_lists[0][0]

    # fmt:off
    workflow.connect([
        (unwarp_wf, join_echos, [
            ("outputnode.corrected", "bold_files"),
        ]),
        (unwarp_wf, join_sdc_echos, [
            ("outputnode.fieldmap", "fieldmap"),
            ("outputnode.fieldwarp", "fieldwarp"),
            ("outputnode.corrected", "corrected"),
            ("outputnode.corrected_ref", "corrected_ref"),
            ("outputnode.corrected_mask", "corrected_mask"),
        ]),
        # remaining workflow connections
        (join_sdc_echos, final_boldref_wf, [
            ("corrected", "inputnode.bold_file"),
        ]),
        (join_sdc_echos, bold_t2s_wf, [
            (("corrected_mask", pop_file), "inputnode.bold_mask"),
        ]),
        (join_sdc_echos, bold_t1_trans_wf, [
            # TEMPORARY: For the moment we can't use frame-wise fieldmaps
            (("fieldwarp", _dpop), "inputnode.fieldwarp"),
        ]),
        (join_sdc_echos, bold_std_trans_wf, [
            # TEMPORARY: For the moment we can't use frame-wise fieldmaps
            (("fieldwarp", _dpop), "inputnode.fieldwarp"),
        ]),
    ])
    # fmt:on

    return workflow


def _create_mem_gb(bold_fname):
    bold_size_gb = os.path.getsize(bold_fname) / (1024**3)
    bold_tlen = nb.load(bold_fname).shape[-1]
    mem_gb = {
        "filesize": bold_size_gb,
        "resampled": bold_size_gb * 4,
        "largemem": bold_size_gb * (max(bold_tlen / 100, 1.0) + 4),
    }

    return bold_tlen, mem_gb


def _get_wf_name(bold_fname):
    """
    Derive the workflow name for supplied BOLD file.
    >>> _get_wf_name('/completely/made/up/path/sub-01_task-nback_bold.nii.gz')
    'func_preproc_task_nback_wf'
    >>> _get_wf_name('/completely/made/up/path/sub-01_task-nback_run-01_echo-1_bold.nii.gz')
    'func_preproc_task_nback_run_01_echo_1_wf'
    """
    from nipype.utils.filemanip import split_filename

    fname = split_filename(bold_fname)[1]
    fname_nosub = "_".join(fname.split("_")[1:])
    name = "func_preproc_" + fname_nosub.replace(".", "_").replace(" ", "").replace(
        "-", "_"
    ).replace("_bold", "_wf")

    return name


def _to_join(in_file, join_file):
    """Join two tsv files if the join_file is not ``None``."""
    from niworkflows.interfaces.utility import JoinTSVColumns

    if join_file is None:
        return in_file
    res = JoinTSVColumns(in_file=in_file, join_file=join_file).run()
    return res.outputs.out_file


def get_img_orientation(imgf):
    """Return the image orientation as a string"""
    img = nb.load(imgf)
    return "".join(nb.aff2axcodes(img.affine))


def get_MNIInfant_key(spaces):
    """
    Parse spaces and return matching MNIInfant space.
    """
    key = None
    for space in spaces.references:
        # str formats as <reference.name>:<reference.spec>
        if "MNIInfant" in str(space) and "res-native" not in str(space):
            key = str(space).replace(":", "_")
            break

    if key is None:
        raise KeyError(f"MNIInfant not found in SpatialReferences: {spaces}")
    return key



import migas

from .. import __version__, config


def setup_migas(init_ping: bool = True) -> None:
    """
    Prepare the migas python client to communicate with a migas server.
    If `init_ping` is `True`, send an initial breadcrumb.
    """
    # generate session UUID from generated run UUID
    session_id = None
    if config.execution.run_uuid:
        session_id = config.execution.run_uuid.split('_', 1)[-1]

    migas.setup(session_id=session_id)
    if init_ping:
        # send initial status ping
        ping_migas(status='pending')


def ping_migas(*, status: str) -> dict:
    """
    Communicate with the migas telemetry server. This requires `migas.setup()` to be called.
    """
    res = migas.add_project("nipreps/nibabies", __version__, status=status)
    return res



"""
Subcortical alignment into MNI space
"""

from nipype.interfaces import fsl
from nipype.interfaces import utility as niu
from nipype.pipeline import engine as pe
from niworkflows.engine.workflows import LiterateWorkflow as Workflow
from pkg_resources import resource_filename

from ...interfaces.workbench import VolumeLabelImport


def init_subcortical_rois_wf(*, name="subcortical_rois_wf"):
    """
    Refine segmentations into volumes of expected CIFTI subcortical structures.


    Parameters
    ----------
    name : :obj:`str`
        Name of the workflow

    Inputs
    ------
    MNIInfant_aseg : :obj:`str`
        FreeSurfer's aseg in MNIInfant space

    Outputs
    -------
    MNIInfant_rois : :obj:`str`
        Subcortical ROIs in MNIInfant space
    MNI152_rois : :obj:`str`
        Subcortical ROIs in `MNI152NLin6Asym` space
    """
    from niworkflows.interfaces.nibabel import MapLabels
    from templateflow.api import get as get_template

    # TODO: Implement BOLD refinement once InfantFS outputs subj/mri/wmparc.mgz
    # The code is found at
    # https://github.com/DCAN-Labs/dcan-infant-pipeline/blob/
    # 0e9c2fe32fb4a5032d0a2a3e0905ad97fa52b398/PostFreeSurfer/scripts/
    # FreeSurfer2CaretConvertAndRegisterNonlinear.sh
    # Lines 70-78 & 116-127
    # #
    # For now, just use the aseg

    workflow = Workflow(name=name)
    inputnode = pe.Node(niu.IdentityInterface(fields=["MNIInfant_aseg"]), name="inputnode")
    outputnode = pe.Node(
        niu.IdentityInterface(fields=["MNIInfant_rois", "MNI152_rois"]),
        name="outputnode",
    )
    # Fetch the HCP volumetric template
    tpl_rois = get_template(
        "MNI152NLin6Asym", resolution=2, atlas="HCP", suffix="dseg", raise_empty=True
    )
    outputnode.inputs.MNI152_rois = tpl_rois

    # This will only used for the wmparc in subject space
    # For now, define it and don't run it
    # TODO: Move to TemplateFlow

    # tpl_avgwmparc = resource_filename(
    #     'nibabies', 'data/tpl-MNI152NLin6Asym_res-01_desc-avgwmparc_dseg.nii.gz'
    # )
    # applywarp_tpl = pe.Node(
    #     fsl.ApplyWarp(in_file=tpl_avgwmparc, ref_file=tpl_rois, interp="nn"),
    #     name="applywarp_std"
    # )

    map_labels = pe.Node(
        MapLabels(
            mappings_file=resource_filename("nibabies", "data/FreeSurferLabelRemappings.json")
        ),
        name='map_labels',
    )

    subcortical_labels = resource_filename(
        "nibabies", "data/FreeSurferSubcorticalLabelTableLut.txt"
    )
    refine_bold_rois = pe.Node(
        VolumeLabelImport(label_list_file=subcortical_labels, discard_others=True),
        name="refine_bold_rois",
    )

    # fmt: off
    workflow.connect([
        (inputnode, map_labels, [("MNIInfant_aseg", "in_file")]),
        (map_labels, refine_bold_rois, [("out_file", "in_file")]),
        # (applywarp_tpl, refine_std_rois, [("out_file", "in_file")]),
        (refine_bold_rois, outputnode, [("out_file", "MNIInfant_rois")]),
    ])
    # fmt: on
    return workflow


def init_subcortical_mni_alignment_wf(*, vol_sigma=0.8, name="subcortical_mni_alignment_wf"):
    """
    Align individual subcortical structures into MNI space.

    This is a nipype workflow port of the DCAN infant pipeline:
    https://github.com/DCAN-Labs/dcan-infant-pipeline/blob\
    /master/fMRISurface/scripts/SubcorticalAlign_ROIs.sh


    Parameters
    ----------
    name : :obj:`str`
        Name of the workflow
    vol_sigma : :obj:`float`
        The sigma for the gaussian volume smoothing kernel, in mm

    Inputs
    ------
    MNIInfant_bold : :obj:`str`
        BOLD file in MNI Infant space
    MNIInfant_rois : :obj:`str`
        File containing ROIs in MNI Infant space
    MNI152_rois : :obj:`str`
        File containing ROIs in MNI152NLin6Asym space

    Outputs
    -------
    subcortical_volume : :obj:`str`
        Volume file containing all ROIs individually aligned to standard
    subcortical_labels : :obj:`str`
        Volume file containing all labels
    """
    from niworkflows.engine.workflows import LiterateWorkflow as Workflow
    from niworkflows.interfaces.nibabel import MergeROIs

    from ...interfaces.workbench import (
        CiftiCreateDenseTimeseries,
        CiftiCreateLabel,
        CiftiDilate,
        CiftiResample,
        CiftiSeparate,
        CiftiSmooth,
        VolumeAffineResample,
        VolumeAllLabelsToROIs,
        VolumeLabelExportTable,
    )

    # reuse saved atlas to atlas transform
    atlas_xfm = resource_filename("nibabies", "data/MNIInfant_to_MNI1526NLinAsym.mat")
    inputnode = pe.Node(
        niu.IdentityInterface(fields=["MNIInfant_bold", "MNIInfant_rois", "MNI152_rois"]),
        name="inputnode",
    )
    outputnode = pe.Node(
        niu.IdentityInterface(fields=["subcortical_volume", "subcortical_labels"]),
        name="outputnode",
    )

    applyxfm_atlas = pe.Node(fsl.ApplyXFM(in_matrix_file=atlas_xfm), name="applyxfm_atlas")
    vol_resample = pe.Node(
        VolumeAffineResample(method="ENCLOSING_VOXEL", flirt=True, affine=atlas_xfm),
        name="vol_resample",
    )
    subj_rois = pe.Node(VolumeAllLabelsToROIs(label_map=1), name="subj_rois")
    split_rois = pe.Node(fsl.Split(dimension="t"), name="split_rois")
    atlas_rois = pe.Node(VolumeAllLabelsToROIs(label_map=1), name="atlas_rois")
    split_atlas_rois = pe.Node(fsl.Split(dimension="t"), name="split_atlas_rois")
    atlas_labels = pe.Node(VolumeLabelExportTable(label_map=1), name="atlas_labels")
    parse_labels = pe.Node(
        niu.Function(function=parse_roi_labels, output_names=["structures", "label_ids"]),
        name="parse_labels",
    )

    # The following is wrapped in a for-loop, iterating across each roi
    # Instead, we will use MapNodes and iter across the varying inputs
    roi2atlas = pe.MapNode(
        fsl.FLIRT(
            searchr_x=[-20, 20],
            searchr_y=[-20, 20],
            searchr_z=[-20, 20],
            interp="nearestneighbour",
        ),
        name="roi2atlas",
        iterfield=["in_file", "reference"],
    )
    applyxfm_roi = pe.MapNode(
        fsl.ApplyXFM(interp="spline"),
        iterfield=["reference", "in_matrix_file"],
        name="applyxfm_roi",
        mem_gb=4,
    )
    bold_mask_roi = pe.MapNode(
        fsl.ApplyMask(),
        iterfield=["in_file", "mask_file"],
        name="bold_mask_roi",
    )
    mul_roi = pe.MapNode(
        fsl.BinaryMaths(operation="mul"),
        iterfield=["in_file", "operand_value"],
        name="mul_roi",
    )
    mul_atlas_roi = pe.MapNode(
        fsl.BinaryMaths(operation="mul"),
        iterfield=["in_file", "operand_value"],
        name="mul_atlas_roi",
    )
    vol_label = pe.MapNode(
        VolumeLabelImport(drop_unused_labels=True),
        iterfield=["in_file"],
        name="vol_label",
    )
    vol_atlas_label = pe.MapNode(
        VolumeLabelImport(drop_unused_labels=True),
        iterfield=["in_file"],
        name="vol_atlas_label",
    )
    create_dtseries = pe.MapNode(
        CiftiCreateDenseTimeseries(),
        iterfield=["volume_data", "volume_structure_labels"],
        name="create_dtseries",
    )
    create_label = pe.MapNode(
        CiftiCreateLabel(),
        iterfield=["volume_label", "structure_label_volume"],
        name="create_label",
    )
    dilate = pe.MapNode(
        CiftiDilate(direction="COLUMN", surface_distance=0, volume_distance=10),
        iterfield=["in_file"],
        name="dilate",
    )
    resample = pe.MapNode(
        CiftiResample(
            direction="COLUMN",
            template_direction="COLUMN",
            surface_method="ADAP_BARY_AREA",
            volume_method="CUBIC",
            volume_predilate=10,
        ),
        iterfield=["in_file", "template"],
        name="resample",
    )
    smooth = pe.MapNode(
        CiftiSmooth(direction="COLUMN", fix_zeros_vol=True, sigma_surf=0, sigma_vol=vol_sigma),
        iterfield=["in_file"],
        name="smooth",
    )
    separate = pe.MapNode(
        CiftiSeparate(direction="COLUMN", volume_all_file="volume_all.nii.gz"),
        iterfield=["in_file"],
        name="separate",
    )
    fmt_agg_rois = pe.Node(
        niu.Function(
            function=format_agg_rois,
            output_names=["first_image", "op_files", "op_string"],
        ),
        name="fmt_agg_rois",
    )
    agg_rois = pe.Node(fsl.MultiImageMaths(), name="agg_rois")
    merge_rois = pe.Node(MergeROIs(), name="merge_rois")

    workflow = Workflow(name=name)
    # fmt: off
    workflow.connect([
        (inputnode, applyxfm_atlas, [
            ("MNIInfant_bold", "in_file"),
            ("MNI152_rois", "reference")]),
        (inputnode, vol_resample, [
            ("MNIInfant_rois", "in_file"),
            ("MNIInfant_rois", "flirt_source_volume")]),
        (applyxfm_atlas, vol_resample, [
            ("out_file", "volume_space"),
            ("out_file", "flirt_target_volume")]),
        (inputnode, subj_rois, [("MNIInfant_rois", "in_file")]),
        (inputnode, atlas_rois, [("MNI152_rois", "in_file")]),
        (subj_rois, split_rois, [("out_file", "in_file")]),
        (atlas_rois, split_atlas_rois, [("out_file", "in_file")]),
        (inputnode, atlas_labels, [("MNI152_rois", "in_file")]),
        (atlas_labels, parse_labels, [("out_file", "label_file")]),
        # for loop across ROIs
        (split_rois, roi2atlas, [("out_files", "in_file")]),
        (split_atlas_rois, roi2atlas, [("out_files", "reference")]),
        (inputnode, applyxfm_roi, [("MNIInfant_bold", "in_file")]),
        (split_atlas_rois, applyxfm_roi, [("out_files", "reference")]),
        (roi2atlas, applyxfm_roi, [("out_matrix_file", "in_matrix_file")]),
        (applyxfm_roi, bold_mask_roi, [("out_file", "in_file")]),
        (roi2atlas, bold_mask_roi, [("out_file", "mask_file")]),
        (roi2atlas, mul_roi, [("out_file", "in_file")]),
        (parse_labels, mul_roi, [("label_ids", "operand_value")]),
        (split_atlas_rois, mul_atlas_roi, [("out_files", "in_file")]),
        (parse_labels, mul_atlas_roi, [("label_ids", "operand_value")]),
        (mul_roi, vol_label, [("out_file", "in_file")]),
        (atlas_labels, vol_label, [("out_file", "label_list_file")]),
        (mul_atlas_roi, vol_atlas_label, [("out_file", "in_file")]),
        (atlas_labels, vol_atlas_label, [("out_file", "label_list_file")]),
        (bold_mask_roi, create_dtseries, [("out_file", "volume_data")]),
        (vol_label, create_dtseries, [("out_file", "volume_structure_labels")]),
        (vol_atlas_label, create_label, [
            ("out_file", "volume_label"),
            ("out_file", "structure_label_volume")]),
        (create_dtseries, dilate, [("out_file", "in_file")]),
        (dilate, resample, [("out_file", "in_file")]),
        (create_label, resample, [("out_file", "template")]),
        (resample, smooth, [("out_file", "in_file")]),
        (smooth, separate, [("out_file", "in_file")]),
        # end loop
        (mul_roi, fmt_agg_rois, [("out_file", "rois")]),
        (fmt_agg_rois, agg_rois, [
            ("first_image", "in_file"),
            ("op_files", "operand_files"),
            ("op_string", "op_string")]),
        (separate, merge_rois, [("volume_all_file", "in_files")]),
        (merge_rois, outputnode, [("out_file", "subcortical_volume")]),
        (inputnode, outputnode, [("MNI152_rois", "subcortical_labels")]),
    ])
    # fmt: on
    return workflow


def parse_roi_labels(label_file):
    """
    Parse a label file composed of one or more sets of:
    <labelname>
    <key> <red> <green> <blue> <alpha>

    Return a list of structure names and label keys.

    >>> structs, ids = parse_roi_labels(test_data / "labelfile.txt")
    >>> structs
    ['CEREBELLUM_LEFT', 'THALAMUS_LEFT', 'CAUDATE_LEFT']
    >>> ids
    [8, 10, 11]
    """

    with open(label_file) as fp:
        lines = fp.readlines()
    if len(lines) % 2 == 1:
        raise RuntimeError("Label file is incomplete or invalid")
    structs, label_ids = [], []
    for idx, line in enumerate(lines):
        if idx % 2 == 0:
            structs.append(line.strip())
        else:
            label_ids.append(int(line.split(" ", 1)[0]))
    return structs, label_ids


def format_agg_rois(rois):
    """
    Helper function to format MultiImageMaths command.

    Parameters
    ----------
    rois : `list` of `str`s
        List of files

    Returns
    -------
    first_image
    op_files
    op_string

    """
    return rois[0], rois[1:], ("-add %s " * (len(rois) - 1)).strip()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Utilities to handle BIDS inputs."""
import json
import os
import sys
from dataclasses import dataclass, field
from pathlib import Path
from typing import IO, List, Union


@dataclass
class BOLDGrouping:
    """This class is used to facilitate the grouping of BOLD series."""

    session: Union[str, None]
    pe_dir: str
    readout: float
    multiecho_id: str = None
    files: List[IO] = field(default_factory=list)

    @property
    def name(self) -> str:
        return f"{self.session}-{self.pe_dir}-{self.readout}-{self.multiecho_id}"

    def add_file(self, fl) -> None:
        self.files.append(fl)


def write_bidsignore(deriv_dir):
    # TODO: Port to niworkflows
    bids_ignore = (
        "*.html",
        "logs/",
        "figures/",  # Reports
        "*_xfm.*",  # Unspecified transform files
        "*.surf.gii",  # Unspecified structural outputs
        # Unspecified functional outputs
        "*_boldref.nii.gz",
        "*_bold.func.gii",
        "*_mixing.tsv",
        "*_AROMAnoiseICs.csv",
        "*_timeseries.tsv",
    )
    ignore_file = Path(deriv_dir) / ".bidsignore"

    ignore_file.write_text("\n".join(bids_ignore) + "\n")


def write_derivative_description(bids_dir, deriv_dir):
    from ..__about__ import DOWNLOAD_URL, __packagename__, __version__

    bids_dir = Path(bids_dir)
    deriv_dir = Path(deriv_dir)
    desc = {
        "Name": "NiBabies: Neuroimaging preprocessing workflows for babies",
        "BIDSVersion": "1.4.0",
        "DatasetType": "derivative",
        "GeneratedBy": [
            {
                "Name": __packagename__,
                "Version": __version__,
                "CodeURL": DOWNLOAD_URL,
            }
        ],
        "HowToAcknowledge": "TODO",
    }

    # Keys that can only be set by environment
    if "NIBABIES_DOCKER_TAG" in os.environ:
        desc["GeneratedBy"][0]["Container"] = {
            "Type": "docker",
            "Tag": f"nipreps/nibabies:{os.environ['NIBABIES_DOCKER_TAG']}",
        }
    if "NIBABIES_SINGULARITY_URL" in os.environ:
        desc["GeneratedBy"][0]["Container"] = {
            "Type": "singularity",
            "URI": os.getenv("NIBABIES_SINGULARITY_URL"),
        }

    # Keys deriving from source dataset
    orig_desc = {}
    fname = bids_dir / "dataset_description.json"
    if fname.exists():
        orig_desc = json.loads(fname.read_text())

    if "DatasetDOI" in orig_desc:
        desc["SourceDatasets"] = [
            {"URL": f'https://doi.org/{orig_desc["DatasetDOI"]}', "DOI": orig_desc["DatasetDOI"]}
        ]
    if "License" in orig_desc:
        desc["License"] = orig_desc["License"]

    Path.write_text(deriv_dir / "dataset_description.json", json.dumps(desc, indent=4))


def extract_entities(file_list):
    """
    Return a dictionary of common entities given a list of files.
    Examples
    --------
    >>> extract_entities('sub-01/anat/sub-01_T1w.nii.gz')
    {'subject': '01', 'suffix': 'T1w', 'datatype': 'anat', 'extension': '.nii.gz'}
    >>> extract_entities(['sub-01/anat/sub-01_T1w.nii.gz'] * 2)
    {'subject': '01', 'suffix': 'T1w', 'datatype': 'anat', 'extension': '.nii.gz'}
    >>> extract_entities(['sub-01/anat/sub-01_run-1_T1w.nii.gz',
    ...                   'sub-01/anat/sub-01_run-2_T1w.nii.gz'])
    {'subject': '01', 'run': [1, 2], 'suffix': 'T1w', 'datatype': 'anat', 'extension': '.nii.gz'}
    """
    from collections import defaultdict

    from bids.layout import parse_file_entities
    from niworkflows.utils.connections import listify

    entities = defaultdict(list)
    for e, v in [
        ev_pair for f in listify(file_list) for ev_pair in parse_file_entities(f).items()
    ]:
        entities[e].append(v)

    def _unique(inlist):
        inlist = sorted(set(inlist))
        if len(inlist) == 1:
            return inlist[0]
        return inlist

    return {k: _unique(v) for k, v in entities.items()}


def group_bolds_ref(*, layout, subject, sessions=None):
    """
    Extracts BOLD files from a BIDS dataset and combines them into buckets.
    Files in a bucket share:
    1) Session
    2) Phase-encoding direction (PEdir)
    3) Total readout time (TRT)

    Any files with missing data for (2) or (3) are put in their own bucket.

    Parameters
    ----------
    layout : pybids.layout.BIDSLayout
        Initialized BIDSLayout
    subject : str
        The subject ID
    sessions : None

    Outputs
    -------
    combinations : list of tuples
        Each tuple is composed of (session, PEdir, TRT)
    files : list of lists
        Files matching each combination.

    Limitations
    -----------
    Single-band reference (sbref) are excluded.
    """
    import re
    from contextlib import suppress
    from itertools import product

    from sdcflows.utils.epimanip import get_trt

    base_entities = {
        "subject": subject,
        "extension": (".nii", ".nii.gz"),
        "scope": "raw",  # Ensure derivatives are not captured
    }
    # dictionary containing unique Groupings and files
    groupings = {}
    # list of all BOLDS encountered
    all_bolds = []

    sessions = sessions if sessions else layout.get_sessions(subject=subject, scope="raw")

    for ses, suffix in sorted(product(sessions or (None,), {"bold"})):
        # bold files same session
        bolds = layout.get(suffix=suffix, session=ses, **base_entities)
        # some sessions may not have BOLD scans
        if bolds is None:
            continue

        for i, bold in enumerate(bolds):
            multiecho_id = None
            # multi-echo should be grouped together
            if 'echo' in bold.entities:
                # create unique id by dropping "_echo-{i}"
                multiecho_id = re.sub(r"_echo-\d+", "", bold.filename)

            # session, pe, ro
            meta = bold.get_metadata()
            pe_dir = meta.get("PhaseEncodingDirection")

            ro = None
            with suppress(ValueError):
                ro = get_trt(meta, bold.path)
            if ro is not None:
                meta.update({"TotalReadoutTime": ro})

            grouping = BOLDGrouping(
                session=ses,
                pe_dir=pe_dir,
                readout=ro,
                multiecho_id=multiecho_id,
            )

            if any(v is None for v in (pe_dir, ro)):
                # cannot be certain so treat as unique
                grouping.add_file(bold.path)
                groupings[f'unknown{i}'] = grouping
            else:
                try:
                    grouping = groupings[grouping.name]
                except KeyError:
                    groupings[grouping.name] = grouping

                grouping.add_file(bold.path)

        all_bolds += bolds

        if len(all_bolds) != sum([len(g.files) for _, g in groupings.items()]):
            msg = f"""Error encountered when grouping BOLD runs.
Combinations: {groupings}
BOLD files: {bolds}

Please file a bug-report with the nibabies developers at:
https://github.com/nipreps/nibabies/issues/new/choose
"""
            raise RuntimeError(msg)

    return groupings


def validate_input_dir(exec_env, bids_dir, participant_label):
    # Ignore issues and warnings that should not influence NiBabies
    import subprocess
    import tempfile

    validator_config_dict = {
        "ignore": [
            "EVENTS_COLUMN_ONSET",
            "EVENTS_COLUMN_DURATION",
            "TSV_EQUAL_ROWS",
            "TSV_EMPTY_CELL",
            "TSV_IMPROPER_NA",
            "VOLUME_COUNT_MISMATCH",
            "BVAL_MULTIPLE_ROWS",
            "BVEC_NUMBER_ROWS",
            "DWI_MISSING_BVAL",
            "INCONSISTENT_SUBJECTS",
            "INCONSISTENT_PARAMETERS",
            "BVEC_ROW_LENGTH",
            "B_FILE",
            "PARTICIPANT_ID_COLUMN",
            "PARTICIPANT_ID_MISMATCH",
            "TASK_NAME_MUST_DEFINE",
            "PHENOTYPE_SUBJECTS_MISSING",
            "STIMULUS_FILE_MISSING",
            "DWI_MISSING_BVEC",
            "EVENTS_TSV_MISSING",
            "TSV_IMPROPER_NA",
            "ACQTIME_FMT",
            "Participants age 89 or higher",
            "DATASET_DESCRIPTION_JSON_MISSING",
            "FILENAME_COLUMN",
            "WRONG_NEW_LINE",
            "MISSING_TSV_COLUMN_CHANNELS",
            "MISSING_TSV_COLUMN_IEEG_CHANNELS",
            "MISSING_TSV_COLUMN_IEEG_ELECTRODES",
            "UNUSED_STIMULUS",
            "CHANNELS_COLUMN_SFREQ",
            "CHANNELS_COLUMN_LOWCUT",
            "CHANNELS_COLUMN_HIGHCUT",
            "CHANNELS_COLUMN_NOTCH",
            "CUSTOM_COLUMN_WITHOUT_DESCRIPTION",
            "ACQTIME_FMT",
            "SUSPICIOUSLY_LONG_EVENT_DESIGN",
            "SUSPICIOUSLY_SHORT_EVENT_DESIGN",
            "MALFORMED_BVEC",
            "MALFORMED_BVAL",
            "MISSING_TSV_COLUMN_EEG_ELECTRODES",
            "MISSING_SESSION",
        ],
        "error": ["NO_T1W"],
        "ignoredFiles": ["/dataset_description.json", "/participants.tsv"],
    }
    # Limit validation only to data from requested participants
    if participant_label:
        all_subs = set([s.name[4:] for s in bids_dir.glob("sub-*")])
        selected_subs = set([s[4:] if s.startswith("sub-") else s for s in participant_label])
        bad_labels = selected_subs.difference(all_subs)
        if bad_labels:
            error_msg = (
                "Data for requested participant(s) label(s) not found. Could "
                "not find data for participant(s): %s. Please verify the requested "
                "participant labels."
            )
            if exec_env == "docker":
                error_msg += (
                    " This error can be caused by the input data not being "
                    "accessible inside the docker container. Please make sure all "
                    "volumes are mounted properly (see https://docs.docker.com/"
                    "engine/reference/commandline/run/#mount-volume--v---read-only)"
                )
            if exec_env == "singularity":
                error_msg += (
                    " This error can be caused by the input data not being "
                    "accessible inside the singularity container. Please make sure "
                    "all paths are mapped properly (see https://www.sylabs.io/"
                    "guides/3.0/user-guide/bind_paths_and_mounts.html)"
                )
            raise RuntimeError(error_msg % ",".join(bad_labels))

        ignored_subs = all_subs.difference(selected_subs)
        if ignored_subs:
            for sub in ignored_subs:
                validator_config_dict["ignoredFiles"].append("/sub-%s/**" % sub)
    with tempfile.NamedTemporaryFile(mode="w+", suffix=".json") as temp:
        temp.write(json.dumps(validator_config_dict))
        temp.flush()
        try:
            subprocess.check_call(["bids-validator", str(bids_dir), "-c", temp.name])
        except FileNotFoundError:
            print("bids-validator does not appear to be installed", file=sys.stderr)


def collect_precomputed_derivatives(layout, subject_id, derivatives_filters=None):
    """
    Query and collect precomputed derivatives.

    This function is used to determine which workflow steps can be skipped,
    based on the files found.
    """

    deriv_queries = {
        'anat_mask': {
            'datatype': 'anat',
            'desc': 'brain',
            'space': 'orig',
            'suffix': 'mask',
        },
        'anat_aseg': {
            'datatype': 'anat',
            'desc': 'aseg',
            'space': 'orig',
            'suffix': 'dseg',
        },
    }
    if derivatives_filters is not None:
        deriv_queries.update(derivatives_filters)

    derivatives = {}
    for deriv, query in deriv_queries.items():
        res = layout.get(
            scope='derivatives',
            subject=subject_id,
            extension=['.nii', '.nii.gz'],
            return_type="filename",
            **query,
        )
        if not res:
            continue
        if len(res) > 1:  # Some queries may want multiple results
            raise Exception(
                f"When searching for <{deriv}>, found multiple results: {[f.path for f in res]}"
            )
        derivatives[deriv] = res[0]
    return derivatives



"Patch functions to allow 'flexible' connections"


def set_tf_resolution(template, sloppy):
    "'Dynamic' workaround to apply tweaks based on templates"
    if template == "UNCInfant":
        from nipype.interfaces.base import Undefined

        return Undefined
    return sloppy + 1


def set_reg_resolution(template):
    if template == "UNCInfant":
        return None
    from nipype.interfaces.base import Undefined

    return Undefined



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""nibabies setup script """
import sys
from setuptools import setup
import versioneer

# Give setuptools a hint to complain if it's too old a version
# 30.3.0 allows us to put most metadata in setup.cfg
# 30.4.0 gives us options.packages.find
# Should match pyproject.toml
SETUP_REQUIRES = ["setuptools >= 30.4.0"]
# This enables setuptools to install wheel on-the-fly
SETUP_REQUIRES += ["wheel"] if "bdist_wheel" in sys.argv else []

if __name__ == "__main__":
    setup(
        name="nibabies",
        version=versioneer.get_version(),
        cmdclass=versioneer.get_cmdclass(),
        setup_requires=SETUP_REQUIRES,
    )


"""Manipulate Python warnings."""
import logging
import warnings

_wlog = logging.getLogger("py.warnings")
_wlog.addHandler(logging.NullHandler())


def _warn(message, category=None, stacklevel=1, source=None):
    """Redefine the warning function."""
    if category is not None:
        category = type(category).__name__
        category = category.replace("type", "WARNING")

    logging.getLogger("py.warnings").warning(f"{category or 'WARNING'}: {message}")


def _showwarning(message, category, filename, lineno, file=None, line=None):
    _warn(message, category=category)


warnings.warn = _warn
warnings.showwarning = _showwarning



import os

from nipype.interfaces.base import CommandLineInputSpec, File, Str, TraitedSpec, traits
from nipype.interfaces.base.traits_extension import (
    InputMultiObject,
    OutputMultiObject,
    isdefined,
)
from nipype.interfaces.workbench.base import WBCommand

# patch
from nipype.interfaces.workbench.cifti import CiftiSmooth as _CiftiSmooth
from nipype.interfaces.workbench.cifti import (
    CiftiSmoothInputSpec as _CiftiSmoothInputSpec,
)

VALID_STRUCTURES = (
    "CORTEX_LEFT",
    "CORTEX_RIGHT",
    "CEREBELLUM",
    "ACCUMBENS_LEFT",
    "ACCUMBENS_RIGHT",
    "ALL_GREY_MATTER",
    "ALL_WHITE_MATTER",
    "AMYGDALA_LEFT",
    "AMYGDALA_RIGHT",
    "BRAIN_STEM",
    "CAUDATE_LEFT",
    "CAUDATE_RIGHT",
    "CEREBELLAR_WHITE_MATTER_LEFT",
    "CEREBELLAR_WHITE_MATTER_RIGHT",
    "CEREBELLUM_LEFT",
    "CEREBELLUM_RIGHT",
    "CEREBRAL_WHITE_MATTER_LEFT",
    "CEREBRAL_WHITE_MATTER_RIGHT",
    "CORTEX",
    "DIENCEPHALON_VENTRAL_LEFT",
    "DIENCEPHALON_VENTRAL_RIGHT",
    "HIPPOCAMPUS_LEFT",
    "HIPPOCAMPUS_RIGHT",
    "INVALID",
    "OTHER",
    "OTHER_GREY_MATTER",
    "OTHER_WHITE_MATTER",
    "PALLIDUM_LEFT",
    "PALLIDUM_RIGHT",
    "PUTAMEN_LEFT",
    "PUTAMEN_RIGHT",
    "THALAMUS_LEFT",
    "THALAMUS_RIGHT",
)


class CiftiCreateDenseFromTemplateInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=0,
        desc="File to match brainordinates of",
    )
    out_file = File(
        name_source=["in_file"],
        name_template="template_%s.nii",
        keep_extension=True,
        argstr="%s",
        position=1,
        desc="The output CIFTI file",
    )
    series = traits.Bool(
        argstr="-series",
        position=2,
        desc="Make a dtseries file instead of a dscalar",
    )
    series_step = traits.Float(
        requires=["series"],
        argstr="%.1f",
        position=3,
        desc="Increment between series points",
    )
    series_start = traits.Float(
        requires=["series"],
        argstr="%.1f",
        position=4,
        desc="Start value of the series",
    )
    series_unit = traits.Enum(
        "SECOND",
        "HERTZ",
        "METER",
        "RADIAN",
        requries=["series"],
        argstr="-unit %s",
        position=5,
        desc="select unit for series (default SECOND)",
    )
    volume_all = traits.File(
        exists=True,
        argstr="-volume-all %s",
        position=6,
        desc="the input volume file for all voxel data",
    )
    volume_all_from_cropped = traits.Bool(
        requires=["volume_all"],
        argstr="-from-cropped",
        position=7,
        desc="the input is cropped to the size of the voxel data in the template file",
    )
    label_collision = traits.Enum(
        "ERROR",
        "SURFACES_FIRST",
        "LEGACY",
        argstr="-label-collision %s",
        position=8,
        desc="how to handle conflicts between label keys, use 'LEGACY' to match v1.4.2 "
        "and earlier",
    )
    cifti = InputMultiObject(
        File(exists=True),
        argstr="-cifti %s",
        position=9,
        desc="use input data from one or more CIFTI files",
    )
    metric = InputMultiObject(
        traits.Tuple(traits.Enum(VALID_STRUCTURES), File(exists=True)),
        argstr="%s",
        position=10,
        desc="use input data from one or more metric files",
    )
    label = InputMultiObject(
        traits.Tuple(traits.Enum(VALID_STRUCTURES), File(exists=True)),
        argstr="%s",
        position=11,
        desc="use input data from one or more surface label files",
    )
    volume = InputMultiObject(
        traits.Either(
            traits.Tuple(traits.Enum(VALID_STRUCTURES), File(exists=True)),
            traits.Tuple(traits.Enum(VALID_STRUCTURES), File(exists=True), traits.Bool()),
        ),
        argstr="%s",
        position=12,
        desc="use a volume file for a single volume structure's data",
    )


class CiftiCreateDenseFromTemplateOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="The output CIFTI file")


class CiftiCreateDenseFromTemplate(WBCommand):
    """
    This command helps you make a new dscalar, dtseries, or dlabel cifti file
    that matches the brainordinate space used in another cifti file.  The
    template file must have the desired brainordinate space in the mapping
    along the column direction (for dtseries, dscalar, dlabel, and symmetric
    dconn this is always the case).  All input cifti files must have a brain
    models mapping along column and use the same volume space and/or surface
    vertex count as the template for structures that they contain.  If any
    input files contain label data, then input files with non-label data are
    not allowed, and the -series option may not be used.

    Any structure that isn't covered by an input is filled with zeros or the
    unlabeled key.

    >>> from nibabies.interfaces import workbench as wb
    >>> frmtpl = wb.CiftiCreateDenseFromTemplate()
    >>> frmtpl.inputs.in_file = data_dir / "func.dtseries.nii"
    >>> frmtpl.inputs.series = True
    >>> frmtpl.inputs.series_step = 0.8
    >>> frmtpl.inputs.series_start = 0
    >>> frmtpl.cmdline
    'wb_command -cifti-create-dense-from-template .../func.dtseries.nii \
    template_func.dtseries.nii -series 0.8 0.0'

    >>> frmtpl.inputs.volume = [("OTHER", data_dir / 'functional.nii', True), \
        ("PUTAMEN_LEFT", data_dir / 'functional.nii')]
    >>> frmtpl.cmdline
    'wb_command -cifti-create-dense-from-template .../func.dtseries.nii \
    template_func.dtseries.nii -series 0.8 0.0 \
    -volume OTHER .../functional.nii -from-cropped \
    -volume PUTAMEN_LEFT .../functional.nii'
    """

    input_spec = CiftiCreateDenseFromTemplateInputSpec
    output_spec = CiftiCreateDenseFromTemplateOutputSpec
    _cmd = "wb_command -cifti-create-dense-from-template"

    def _format_arg(self, name, trait_spec, value):
        if name in ("metric", "label", "volume"):
            cmds = []
            for val in value:
                if val[-1] is True:  # volume specific
                    val = val[:2] + ("-from-cropped",)
                cmds.append(" ".join((f"-{name}",) + val))
            return trait_spec.argstr % " ".join(cmds)
        return super()._format_arg(name, trait_spec, value)


class CiftiCreateDenseTimeseriesInputSpec(CommandLineInputSpec):
    out_file = File(
        value="out.dtseries.nii",
        usedefault=True,
        argstr="%s",
        position=0,
        desc="The output CIFTI file",
    )
    volume_data = File(
        exists=True,
        argstr="-volume %s",
        position=1,
        requires=["volume_structure_labels"],
        desc="volume file containing all voxel data for all volume structures",
    )
    volume_structure_labels = File(
        exists=True,
        argstr="%s",
        position=2,
        requires=["volume_data"],
        desc="label volume file containing labels for cifti structures",
    )
    left_metric = File(
        exists=True,
        argstr="-left-metric %s",
        position=3,
        desc="metric file for left surface",
    )
    roi_left = File(
        exists=True,
        argstr="-roi-left %s",
        position=4,
        requires=["left_metric"],
        desc="ROI (as metric file) of vertices to use from left surface",
    )
    right_metric = File(
        exists=True,
        argstr="-right-metric %s",
        position=5,
        desc="metric file for right surface",
    )
    roi_right = File(
        exists=True,
        argstr="-roi-right %s",
        position=6,
        requires=["right_metric"],
        desc="ROI (as metric file) of vertices to use from right surface",
    )
    cerebellum_metric = File(
        exists=True,
        argstr="-cerebellum-metric %s",
        position=7,
        desc="metric file for cerebellum",
    )
    roi_cerebellum = File(
        exists=True,
        argstr="-roi-cerebellum %s",
        position=8,
        requires=["cerebellum_metric"],
        desc="ROI (as metric file) of vertices to use from cerebellum",
    )
    timestart = traits.Float(
        argstr="-timestart %g",
        position=9,
        desc="the time at the first frame, in seconds",
    )
    timestep = traits.Float(
        argstr="-timestep %g",
        position=10,
        desc="the timestep, in seconds",
    )
    unit = traits.Enum(
        "SECOND",
        "HERTZ",
        "METER",
        "RADIAN",
        argstr="-unit %s",
        position=11,
        desc="use a unit other than time",
    )


class CiftiCreateDenseTimeseriesOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="CIFTI dense timeseries file")


class CiftiCreateDenseTimeseries(WBCommand):
    """
    Create a CIFTI dense timeseries.

    All input files must have the same number of columns/subvolumes.  Only
    the specified components will be in the output cifti.  At least one
    component must be specified.

    See -volume-label-import and -volume-help for format details of label
    volume files.  The structure-label-volume should have some of the label
    names from this list, all other label names will be ignored:

    CORTEX_LEFT
    CORTEX_RIGHT
    CEREBELLUM
    ACCUMBENS_LEFT
    ACCUMBENS_RIGHT
    ALL_GREY_MATTER
    ALL_WHITE_MATTER
    AMYGDALA_LEFT
    AMYGDALA_RIGHT
    BRAIN_STEM
    CAUDATE_LEFT
    CAUDATE_RIGHT
    CEREBELLAR_WHITE_MATTER_LEFT
    CEREBELLAR_WHITE_MATTER_RIGHT
    CEREBELLUM_LEFT
    CEREBELLUM_RIGHT
    CEREBRAL_WHITE_MATTER_LEFT
    CEREBRAL_WHITE_MATTER_RIGHT
    CORTEX
    DIENCEPHALON_VENTRAL_LEFT
    DIENCEPHALON_VENTRAL_RIGHT
    HIPPOCAMPUS_LEFT
    HIPPOCAMPUS_RIGHT
    INVALID
    OTHER
    OTHER_GREY_MATTER
    OTHER_WHITE_MATTER
    PALLIDUM_LEFT
    PALLIDUM_RIGHT
    PUTAMEN_LEFT
    PUTAMEN_RIGHT
    THALAMUS_LEFT
    THALAMUS_RIGHT

    >>> from nibabies.interfaces.workbench import CiftiCreateDenseTimeseries
    >>> createdts = CiftiCreateDenseTimeseries()
    >>> createdts.inputs.volume_data = data_dir /'functional.nii'
    >>> createdts.inputs.volume_structure_labels = data_dir / 'atlas.nii'
    >>> createdts.cmdline
    'wb_command -cifti-create-dense-timeseries out.dtseries.nii \
    -volume .../functional.nii .../atlas.nii'
    """

    input_spec = CiftiCreateDenseTimeseriesInputSpec
    output_spec = CiftiCreateDenseTimeseriesOutputSpec
    _cmd = "wb_command -cifti-create-dense-timeseries"

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = os.path.abspath(self.inputs.out_file)
        return outputs


class CiftiCreateLabelInputSpec(CommandLineInputSpec):
    out_file = File(
        value="out.dlabel.nii",
        usedefault=True,
        argstr="%s",
        position=0,
        desc="the output CIFTI file",
    )
    volume_label = File(
        exists=True,
        requires=["structure_label_volume"],
        argstr="-volume %s",
        position=1,
        desc="label volume file containing the data to be copied",
    )
    structure_label_volume = File(
        exists=True,
        requires=["volume_label"],
        argstr="%s",
        position=2,
        desc="label volume file that defines which voxels to use",
    )
    left_label = File(
        exists=True,
        argstr="-left-label %s",
        position=3,
        desc="Label file for left surface",
    )
    left_roi = File(
        exists=True,
        requires=["left_label"],
        argstr="-roi-left %s",
        position=4,
        desc="roi of vertices to use from left surface as a metric file",
    )
    right_label = File(
        exists=True,
        argstr="-right-label %s",
        position=5,
        desc="Label file for right surface",
    )
    right_roi = File(
        exists=True,
        requires=["right_label"],
        argstr="-roi-right %s",
        position=6,
        desc="roi of vertices to use from right surface as a metric file",
    )
    cerebellum_label = File(
        exists=True,
        argstr="-cerebellum-label %s",
        position=7,
        desc="label for the cerebellum",
    )
    cerebellum_roi = File(
        exists=True,
        requires=["cerebellum_label"],
        argstr="-roi-cerebellum %s",
        position=8,
        desc="roi of vertices to use from cerebellum",
    )


class CiftiCreateLabelOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output CIFTI file")


class CiftiCreateLabel(WBCommand):
    """
    All input files must have the same number of columns/subvolumes.  Only
    the specified components will be in the output cifti.  At least one
    component must be specified.

    The -volume option requires two volume arguments, the label-volume
    argument contains all labels you want to display (e.g. nuclei of the
    thalamus), whereas the structure-label-volume argument contains all CIFTI
    voxel-based structures you want to include data within (e.g.
    THALAMUS_LEFT, THALAMUS_RIGHT, etc).  See -volume-label-import and
    -volume-help for format details of label volume files.  If you just want
    the labels in voxels to be the structure names, you may use the same file
    for both arguments.  The structure-label-volume must use some of the
    label names from this list, all other label names in the
    structure-label-volume will be ignored:

    CORTEX_LEFT
    CORTEX_RIGHT
    CEREBELLUM
    ACCUMBENS_LEFT
    ACCUMBENS_RIGHT
    ALL_GREY_MATTER
    ALL_WHITE_MATTER
    AMYGDALA_LEFT
    AMYGDALA_RIGHT
    BRAIN_STEM
    CAUDATE_LEFT
    CAUDATE_RIGHT
    CEREBELLAR_WHITE_MATTER_LEFT
    CEREBELLAR_WHITE_MATTER_RIGHT
    CEREBELLUM_LEFT
    CEREBELLUM_RIGHT
    CEREBRAL_WHITE_MATTER_LEFT
    CEREBRAL_WHITE_MATTER_RIGHT
    CORTEX
    DIENCEPHALON_VENTRAL_LEFT
    DIENCEPHALON_VENTRAL_RIGHT
    HIPPOCAMPUS_LEFT
    HIPPOCAMPUS_RIGHT
    INVALID
    OTHER
    OTHER_GREY_MATTER
    OTHER_WHITE_MATTER
    PALLIDUM_LEFT
    PALLIDUM_RIGHT
    PUTAMEN_LEFT
    PUTAMEN_RIGHT
    THALAMUS_LEFT
    THALAMUS_RIGHT

    >>> from nibabies.interfaces import workbench as wb
    >>> lab = wb.CiftiCreateLabel()
    >>> lab.inputs.volume_label = data_dir / "functional.nii"
    >>> lab.inputs.structure_label_volume = data_dir / "functional.nii"
    >>> lab.cmdline
    'wb_command -cifti-create-label out.dlabel.nii -volume .../functional.nii .../functional.nii'
    """

    input_spec = CiftiCreateLabelInputSpec
    output_spec = CiftiCreateLabelOutputSpec
    _cmd = "wb_command -cifti-create-label"

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = os.path.abspath(self.inputs.out_file)
        return outputs


class CiftiDilateInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=0,
        desc="The input CIFTI file",
    )
    direction = traits.Enum(
        "ROW",
        "COLUMN",
        mandatory=True,
        argstr="%s",
        position=1,
        desc="Which dimension to dilate along, ROW or COLUMN",
    )
    surface_distance = traits.Int(
        mandatory=True,
        argstr="%d",
        position=2,
        desc="The distance to dilate on surfaces, in mm",
    )
    volume_distance = traits.Int(
        mandatory=True,
        argstr="%d",
        position=3,
        desc="The distance to dilate in the volume, in mm",
    )
    out_file = File(
        name_source=["in_file"],
        name_template="dilated_%s.nii",
        keep_extension=True,
        argstr="%s",
        position=4,
        desc="The dilated CIFTI file",
    )
    left_surface = File(
        exists=True,
        position=5,
        argstr="-left-surface %s",
        desc="Specify the left surface to use",
    )
    left_corrected_areas = File(
        exists=True,
        position=6,
        requires=["left_surface"],
        argstr="-left-corrected-areas %s",
        desc="vertex areas (as a metric) to use instead of computing them from the left surface.",
    )
    right_surface = File(
        exists=True,
        position=7,
        argstr="-right-surface %s",
        desc="Specify the right surface to use",
    )
    right_corrected_areas = File(
        exists=True,
        position=8,
        requires=["right_surface"],
        argstr="-right-corrected-areas %s",
        desc="vertex areas (as a metric) to use instead of computing them from the right surface",
    )
    cerebellum_surface = File(
        exists=True,
        position=9,
        argstr="-cerebellum-surface %s",
        desc="specify the cerebellum surface to use",
    )
    cerebellum_corrected_areas = File(
        exists=True,
        position=10,
        requires=["cerebellum_surface"],
        argstr="-cerebellum-corrected-areas %s",
        desc="vertex areas (as a metric) to use instead of computing them from the cerebellum "
        "surface",
    )
    bad_brainordinate_roi = File(
        exists=True,
        position=11,
        argstr="-bad-brainordinate-roi %s",
        desc="CIFTI dscalar or dtseries file, positive values denote brainordinates to have their "
        "values replaced",
    )
    nearest = traits.Bool(
        position=12,
        argstr="-nearest",
        desc="Use nearest good value instead of a weighted average",
    )
    merged_volume = traits.Bool(
        position=13,
        argstr="-merged-volume",
        desc="treat volume components as if they were a single component",
    )
    legacy_mode = traits.Bool(
        position=14,
        argstr="-legacy-mode",
        desc="Use the math from v1.3.2 and earlier for weighted dilation",
    )


class CiftiDilateOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="Dilated CIFTI file")


class CiftiDilate(WBCommand):
    """
    Dilate a CIFTI file.

    For all data values designated as bad, if they neighbor a good value or
    are within the specified distance of a good value in the same kind of
    model, replace the value with a distance weighted average of nearby good
    values, otherwise set the value to zero.  If -nearest is specified, it
    will use the value from the closest good value within range instead of a
    weighted average.  When the input file contains label data, nearest
    dilation is used on the surface, and weighted popularity is used in the
    volume.

    The -*-corrected-areas options are intended for dilating on group average
    surfaces, but it is only an approximate correction for the reduction of
    structure in a group average surface.

    If -bad-brainordinate-roi is specified, all values, including those with
    value zero, are good, except for locations with a positive value in the
    ROI.  If it is not specified, only values equal to zero are bad.

    """

    input_spec = CiftiDilateInputSpec
    output_spec = CiftiDilateOutputSpec
    _cmd = "wb_command -cifti-dilate"


class CiftiResampleInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=0,
        desc="the CIFTI file to resample",
    )
    direction = traits.Enum(
        "ROW",
        "COLUMN",
        mandatory=True,
        argstr="%s",
        position=1,
        desc="the direction of the input that should be resampled",
    )
    template = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=2,
        desc="a CIFTI file containing the CIFTI space to resample to",
    )
    template_direction = traits.Enum(
        "ROW",
        "COLUMN",
        mandatory=True,
        argstr="%s",
        position=3,
        desc="the direction of the template to use as the resampling space",
    )
    surface_method = traits.Enum(
        "ADAP_BARY_AREA",
        "BARYCENTRIC",
        mandatory=True,
        argstr="%s",
        position=4,
        desc="surface resampling method",
    )
    volume_method = traits.Enum(
        "CUBIC",
        "ENCLOSING_VOXEL",
        "TRILINEAR",
        mandatory=True,
        argstr="%s",
        position=5,
        desc="volume interpolation method",
    )
    out_file = File(
        name_source=["in_file"],
        keep_extension=True,
        name_template="resampled_%s.nii",
        argstr="%s",
        position=6,
        desc="the output cifti file",
    )
    surface_largest = traits.Bool(
        argstr="-surface-largest",
        position=7,
        desc="use largest weight instead of weighted average or popularity when doing "
        "surface resampling",
    )
    volume_predilate = traits.Int(
        argstr="-volume-predilate %d",
        position=8,
        desc="distance, in mm, to dilate the volume components before resampling",
    )
    volume_predilate_nearest = traits.Bool(
        requires=["volume_predilate"],
        xor=["volume_predilate_weighted"],
        argstr="-nearest",
        position=9,
        desc="use nearest value dilation",
    )
    volume_predilate_weighted = traits.Bool(
        requires=["volume_predilate"],
        xor=["volume_predilate_nearest"],
        argstr="-weighted",
        position=10,
        desc="use weighted dilation (default)",
    )
    volume_predilate_weighted_exponent = traits.Int(
        requires=["volume_predilate_weighted"],
        argstr="-exponent %d",
        position=11,
        desc="exponent 'n' to use in (1 / (distance ^ n)) as the weighting function (default 7)",
    )
    volume_predilate_weighted_legacy = traits.Bool(
        requires=["volume_predilate_weighted"],
        argstr="-legacy-cutoff",
        position=12,
        desc="use v1.3.2 logic for the kernel cutoff",
    )
    surface_postdilate = traits.Int(
        argstr="-surface-postdilate %d",
        position=13,
        desc="distance, in mm, to dilate the surface components after resampling",
    )
    surface_postdilate_nearest = traits.Bool(
        requires=["surface_postdilate"],
        xor=["surface_postdilate_weighted", "surface_postdilate_linear"],
        argstr="-nearest",
        position=14,
        desc="use nearest value dilation",
    )
    surface_postdilate_linear = traits.Bool(
        requires=["surface_postdilate"],
        xor=["surface_postdilate_weighted", "surface_postdilate_nearest"],
        argstr="-linear",
        position=15,
        desc="use linear dilation",
    )
    surface_postdilate_weighted = traits.Bool(
        requires=["surface_postdilate"],
        xor=["surface_postdilate_nearest", "surface_postdilate_linear"],
        argstr="-weighted",
        position=16,
        desc="use weighted dilation (default for non-label data)",
    )
    surface_postdilate_weighted_exponent = traits.Int(
        requires=["surface_postdilate_weighted"],
        argstr="-exponent %d",
        position=17,
        desc="exponent 'n' to use in (area / (distance ^ n)) as the weighting function "
        "(default 6)",
    )
    surface_postdilate_weighted_legacy = traits.Bool(
        requires=["surface_postdilate_weighted"],
        argstr="-legacy-cutoff",
        position=18,
        desc="use v1.3.2 logic for the kernel cutoff",
    )
    affine = File(
        exists=True,
        argstr="-affine %s",
        position=19,
        desc="affine file for transformation on the volume components",
    )
    affine_flirt_source = File(
        exists=True,
        requires=["affine", "affine_flirt_target"],
        argstr="-flirt %s",
        position=20,
        desc="the source volume used when generating the affine. MUST be used if affine is "
        "a flirt affine",
    )
    affine_flirt_target = File(
        exists=True,
        requires=["affine", "affine_flirt_source"],
        argstr="%s",
        position=21,
        desc="the target volume used when generating the affine. MUST be used if affine is "
        "a flirt affine",
    )
    warpfield = File(
        exists=True,
        argstr="-warpfield %s",
        position=22,
        desc="the warpfield to use on the volume components",
    )
    warpfield_fnirt_source = File(
        exists=True,
        requires=["warpfield"],
        argstr="-fnirt %s",
        position=23,
        desc="the source volume used when generating the warpfield. MUST be used if using "
        "a fnirt warpfield",
    )
    left_sphere_current = File(
        exists=True,
        requires=["left_sphere_new"],
        argstr="-left-spheres %s",
        position=24,
        desc="a sphere with the same mesh as the current left surface",
    )
    left_sphere_new = File(
        exists=True,
        requires=["left_sphere_current"],
        argstr="%s",
        position=25,
        desc="a sphere with the new left mesh that is in register with the current sphere",
    )
    left_area_surf_current = File(
        exists=True,
        requires=["left_sphere_current", "left_area_surf_new"],
        argstr="-left-area-surfs %s",
        position=26,
        desc="a relevant left anatomical surface with current mesh",
    )
    left_area_surf_new = File(
        exists=True,
        requires=["left_sphere_new", "left_area_surf_current"],
        argstr="%s",
        position=27,
        desc="a relevant left anatomical surface with new mesh",
    )
    left_area_metric_current = File(
        exists=True,
        requires=["left_sphere_current", "left_area_metric_new"],
        argstr="-left-area-metrics %s",
        position=28,
        desc="a metric file with vertex areas for the current mesh",
    )
    left_area_metric_new = File(
        exists=True,
        requires=["left_sphere_new", "left_area_metric_current"],
        argstr="%s",
        position=29,
        desc="a metric file with vertex areas for the new mesh",
    )
    right_sphere_current = File(
        exists=True,
        requires=["right_sphere_new"],
        argstr="-right-spheres %s",
        position=30,
        desc="a sphere with the same mesh as the current right surface",
    )
    right_sphere_new = File(
        exists=True,
        requires=["right_sphere_current"],
        argstr="%s",
        position=31,
        desc="a sphere with the new right mesh that is in register with the current sphere",
    )
    right_area_surf_current = File(
        exists=True,
        requires=["right_sphere_current", "right_area_surf_new"],
        argstr="-right-area-surfs %s",
        position=32,
        desc="a relevant right anatomical surface with current mesh",
    )
    right_area_surf_new = File(
        exists=True,
        requires=["right_sphere_new", "right_area_surf_current"],
        argstr="%s",
        position=33,
        desc="a relevant right anatomical surface with new mesh",
    )
    right_area_metric_current = File(
        exists=True,
        requires=["right_sphere_current", "right_area_metric_new"],
        argstr="-right-area-metrics %s",
        position=34,
        desc="a metric file with vertex areas for the current mesh",
    )
    right_area_metric_new = File(
        exists=True,
        requires=["right_sphere_new", "right_area_metric_current"],
        argstr="%s",
        position=35,
        desc="a metric file with vertex areas for the new mesh",
    )
    cerebellum_sphere_current = File(
        exists=True,
        requires=["cerebellum_sphere_new"],
        argstr="-cerebellum-spheres %s",
        position=36,
        desc="a sphere with the same mesh as the current cerebellum surface",
    )
    cerebellum_sphere_new = File(
        exists=True,
        requires=["cerebellum_sphere_current"],
        argstr="%s",
        position=37,
        desc="a sphere with the new cerebellum mesh that is in register with the current sphere",
    )
    cerebellum_area_surf_current = File(
        exists=True,
        requires=["cerebellum_sphere_current", "cerebellum_area_surf_new"],
        argstr="-cerebellum-area-surfs %s",
        position=38,
        desc="a relevant cerebellum anatomical surface with current mesh",
    )
    cerebellum_area_surf_new = File(
        exists=True,
        requires=["cerebellum_sphere_new", "cerebellum_area_surf_current"],
        argstr="%s",
        position=39,
        desc="a relevant cerebellum anatomical surface with new mesh",
    )
    cerebellum_area_metric_current = File(
        exists=True,
        requires=["cerebellum_sphere_current", "cerebellum_area_metric_new"],
        argstr="-cerebellum-area-metrics %s",
        position=40,
        desc="a metric file with vertex areas for the current mesh",
    )
    cerebellum_area_metric_new = File(
        exists=True,
        requires=["cerebellum_sphere_new", "cerebellum_area_metric_current"],
        argstr="%s",
        position=41,
        desc="a metric file with vertex areas for the new mesh",
    )


class CiftiResampleOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the resampled CIFTI file")


class CiftiResample(WBCommand):
    """
    Resample cifti data to a different brainordinate space.  Use COLUMN for
    the direction to resample dscalar, dlabel, or dtseries.  Resampling both
    dimensions of a dconn requires running this command twice, once with
    COLUMN and once with ROW.  If you are resampling a dconn and your machine
    has a large amount of memory, you might consider using
    -cifti-resample-dconn-memory to avoid writing and rereading an
    intermediate file.  The <template-direction> argument should usually be
    COLUMN, as dtseries, dscalar, and dlabel all have brainordinates on that
    direction.  If spheres are not specified for a surface structure which
    exists in the cifti files, its data is copied without resampling or
    dilation.  Dilation is done with the 'nearest' method, and is done on
    <new-sphere> for surface data.  Volume components are padded before
    dilation so that dilation doesn't run into the edge of the component
    bounding box.  If neither -affine nor -warpfield are specified, the
    identity transform is assumed for the volume data.

    The recommended resampling methods are ADAP_BARY_AREA and CUBIC (cubic
    spline), except for label data which should use ADAP_BARY_AREA and
    ENCLOSING_VOXEL.  Using ADAP_BARY_AREA requires specifying an area option
    to each used -*-spheres option.

    >>> from nibabies.interfaces import workbench as wb
    >>> res = wb.CiftiResample()
    >>> res.inputs.in_file = data_dir / "func.dtseries.nii"
    >>> res.inputs.direction = "COLUMN"
    >>> res.inputs.template = data_dir / "func.dlabel.nii"
    >>> res.inputs.template_direction = "COLUMN"
    >>> res.inputs.surface_method = "ADAP_BARY_AREA"
    >>> res.inputs.volume_method = "CUBIC"
    >>> res.inputs.out_file = "resampled.dtseries.nii"
    >>> res.inputs.volume_predilate = 10
    >>> res.cmdline
    'wb_command -cifti-resample .../func.dtseries.nii COLUMN .../func.dlabel.nii COLUMN \
    ADAP_BARY_AREA CUBIC resampled.dtseries.nii -volume-predilate 10'
    """

    input_spec = CiftiResampleInputSpec
    output_spec = CiftiResampleOutputSpec
    _cmd = "wb_command -cifti-resample"


class CiftiSeparateInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=0,
        desc="the cifti to separate a component of",
    )
    direction = traits.Enum(
        "ROW",
        "COLUMN",
        mandatory=True,
        argstr="%s",
        position=1,
        desc="which dimension to smooth along, ROW or COLUMN",
    )
    volume_all_file = File(
        argstr="-volume-all %s",
        position=2,
        desc="separate all volume structures into a volume file",
    )
    volume_all_roi_file = File(
        argstr="-roi %s",
        position=3,
        requires=["volume_all_file"],
        desc="output the roi of which voxels have data",
    )
    volume_all_label_file = File(
        argstr="-label %s",
        position=4,
        requires=["volume_all_file"],
        desc="output a volume label file indicating the location of structures",
    )
    volume_all_crop = traits.Bool(
        argstr="-crop",
        position=5,
        requires=["volume_all_file"],
        desc="crop volume to the size of the data rather than using the original volume size",
    )
    # the following can be repeated
    label = InputMultiObject(
        traits.Either(
            traits.Tuple(traits.Enum(VALID_STRUCTURES), File()),
            traits.Tuple(traits.Enum(VALID_STRUCTURES), File(), File()),
        ),
        argstr="%s",
        position=6,
        desc="separate one or more surface models into a surface label file",
    )
    metric = InputMultiObject(
        traits.Either(
            traits.Tuple(traits.Enum(VALID_STRUCTURES), File()),
            traits.Tuple(traits.Enum(VALID_STRUCTURES), File(), File()),  # -roi
        ),
        argstr="%s",
        position=7,
        desc="separate one or more surface models into a metric file",
    )
    volume = InputMultiObject(
        traits.Either(
            traits.Tuple(traits.Enum(VALID_STRUCTURES), File()),
            traits.Tuple(traits.Enum(VALID_STRUCTURES), File(), File()),  # -roi
            traits.Tuple(traits.Enum(VALID_STRUCTURES, File(), traits.Bool)),  # -crop
            traits.Tuple(traits.Enum(VALID_STRUCTURES), File(), File(), traits.Bool),  # -roi -crop
        ),
        argstr="%s",
        position=8,
        desc="separate one or more volume structures into a volume file",
    )


class CiftiSeparateOutputSpec(TraitedSpec):
    volume_all_file = File(desc="File containing all volume structures")
    volume_all_roi_file = File(desc="Output the roi of which voxels have data")
    volume_all_label_file = File(
        desc="output a volume label file indicating the location of structures"
    )
    label_files = OutputMultiObject(File(), desc="Output label files")
    label_roi_files = OutputMultiObject(File(), desc="Output label rois files")
    metric_files = OutputMultiObject(File(), desc="Output metric files")
    metric_roi_files = OutputMultiObject(File(), desc="Output metric rois files")
    volume_files = OutputMultiObject(File(), desc="Output volume files")
    volume_roi_files = OutputMultiObject(File(), desc="Output volume roi files")


class CiftiSeparate(WBCommand):
    """
    Extract left or right hemisphere surface from CIFTI file (.dtseries)
    other structure can also be extracted
    The input cifti file must have a brain models mapping on the chosen
    dimension, columns for .dtseries.

    >>> separate = CiftiSeparate()
    >>> separate.inputs.in_file = data_dir / "func.dtseries.nii"
    >>> separate.inputs.direction = "COLUMN"
    >>> separate.inputs.volume_all_file = "volume_all.nii.gz"
    >>> separate.cmdline
    'wb_command -cifti-separate .../func.dtseries.nii COLUMN \
    -volume-all volume_all.nii.gz'

    Metrics, labels, and volumes can also be freely extracted
    >>> separate.inputs.metric = [("CORTEX_LEFT", "cortexleft.func.gii")]
    >>> separate.inputs.volume = [("HIPPOCAMPUS_LEFT", "hippoL.nii.gz"), \
        ("HIPPOCAMPUS_RIGHT", "hippoR.nii.gz", "hippoR.roi.nii.gz")]
    >>> separate.cmdline
    'wb_command -cifti-separate .../func.dtseries.nii COLUMN \
    -volume-all volume_all.nii.gz -metric CORTEX_LEFT cortexleft.func.gii \
    -volume HIPPOCAMPUS_LEFT hippoL.nii.gz \
    -volume HIPPOCAMPUS_RIGHT hippoR.nii.gz -roi hippoR.roi.nii.gz'

    """

    input_spec = CiftiSeparateInputSpec
    output_spec = CiftiSeparateOutputSpec
    _cmd = "wb_command -cifti-separate"
    _label_roi_files = []
    _metric_roi_files = []
    _volume_roi_files = []

    def _format_arg(self, name, trait_spec, value):
        if name in ("label", "metric", "volume"):
            cmds = []
            for i, val in enumerate(value):
                if len(val) == 3:
                    if val[-1] is True:
                        val = val[:-1] + ("-crop",)
                    else:
                        val = val[:-1] + ("-roi", val[-1])
                        self._set_roi_file(name, val[-1])
                elif len(val) == 4:
                    val = val[:-2] + ("-roi", val[-2]) + ("crop") if val[-1] is True else ()
                    self._set_roi_file(name, val[-2])
                cmds.append(" ".join((f"-{name}",) + val))
            return trait_spec.argstr % " ".join(cmds)
        return super()._format_arg(name, trait_spec, value)

    def _list_outputs(self):
        outputs = self.output_spec().get()
        if self.inputs.volume_all_file:
            outputs["volume_all_file"] = os.path.abspath(self.inputs.volume_all_file)
        if self.inputs.volume_all_roi_file:
            outputs["volume_all_roi_file"] = os.path.abspath(self.inputs.volume_all_roi_file)
        if self.inputs.volume_all_label_file:
            outputs["volume_all_label_file"] = os.path.abspath(self.inputs.volume_all_label_file)
        if self.inputs.label:
            for label in self.inputs.label:
                outputs["label_files"] = (outputs["label_files"] or []) + self._gen_filename(
                    label[2]
                )
            if self._label_roi_files:
                outputs["label_roi_files"] = self._label_roi_files
        if self.inputs.metric:
            for metric in self.inputs.metric:
                outputs["metric_files"] = (outputs["metric_files"] or []) + self._gen_filename(
                    metric[2]
                )
            if self._metric_roi_files:
                outputs["metric_roi_files"] = self._metric_roi_files
        if self.inputs.volume:
            for volume in self.inputs.volume:
                outputs["volume_files"] = (outputs["volume_files"] or []) + self._gen_filename(
                    volume[2]
                )
            if self._volume_roi_files:
                outputs["volume_roi_files"] = self._volume_roi_files
        return outputs

    def _set_roi_file(self, name, file):
        rois = getattr(self, f"_{name}_roi_files")
        rois.append(self._gen_filename(file))


class CiftiSmoothInputSpec(_CiftiSmoothInputSpec):
    left_surf = File(
        exists=True,
        position=5,
        argstr="-left-surface %s",
        desc="Specify the left surface to use",
    )
    right_surf = File(
        exists=True,
        position=7,
        argstr="-right-surface %s",
        desc="Specify the right surface to use",
    )


class CiftiSmooth(_CiftiSmooth):
    input_spec = CiftiSmoothInputSpec


class VolumeAffineResampleInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=0,
        desc="volume to resample",
    )
    volume_space = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=1,
        desc="a volume file in the volume space you want for the output",
    )
    method = traits.Enum(
        "CUBIC",
        "ENCLOSING_VOXEL",
        "TRILINEAR",
        mandatory=True,
        argstr="%s",
        position=2,
        desc="The resampling method. The recommended methods are CUBIC "
        "(cubic spline) for most data, and ENCLOSING_VOXEL for label data.",
    )
    out_file = File(
        name_source=["in_file"],
        name_template="resampled_%s.nii.gz",
        keep_extension=True,
        argstr="%s",
        position=3,
        desc="the output volume",
    )
    affine = File(
        exists=True,
        mandatory=True,
        argstr="-affine %s",
        position=4,
        desc="the affine file to apply",
    )
    flirt = traits.Bool(
        argstr="-flirt %s %s",
        position=5,
        desc="Set ``True`` if ``affine`` is a FLIRT affine.",
    )
    flirt_source_volume = File(
        exists=True,
        desc="the source volume used when generating the affine; defaults to in_file",
        requires=["flirt"],
    )
    flirt_target_volume = File(
        exists=True,
        desc="the target volume used when generating the affine; defaults to volume_space",
        requires=["flirt"],
    )


class VolumeAffineResampleOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output volume")


class VolumeAffineResample(WBCommand):
    """
    Resample a volume file with an affine transformation.

    >>> from nibabies.interfaces.workbench import VolumeAffineResample
    >>> resample = VolumeAffineResample()
    >>> resample.inputs.in_file = data_dir /'functional.nii'
    >>> resample.inputs.volume_space = data_dir /'anatomical.nii'
    >>> resample.inputs.method = 'CUBIC'
    >>> resample.inputs.affine = data_dir / 'func_to_struct.mat'
    >>> resample.cmdline
    'wb_command -volume-resample .../functional.nii .../anatomical.nii CUBIC \
    resampled_functional.nii.gz -affine .../func_to_struct.mat'

    If the affine was generated with FLIRT, this should be indicated.
    By default, the interface will use the ``in_file`` and ``volume_space``
    for references.

    >>> resample.inputs.flirt = True
    >>> resample.cmdline
    'wb_command -volume-resample .../functional.nii .../anatomical.nii CUBIC \
    resampled_functional.nii.gz -affine .../func_to_struct.mat \
    -flirt .../functional.nii .../anatomical.nii'

    However, if other volumes were used to calculate the affine, they can
    be provided:

    >>> resample.inputs.flirt_source_volume = data_dir / 'epi.nii'
    >>> resample.inputs.flirt_target_volume = data_dir /'T1w.nii'
    >>> resample.cmdline
    'wb_command -volume-resample .../functional.nii .../anatomical.nii CUBIC \
    resampled_functional.nii.gz -affine .../func_to_struct.mat \
    -flirt .../epi.nii .../T1w.nii'
    """

    input_spec = VolumeAffineResampleInputSpec
    output_spec = VolumeAffineResampleOutputSpec
    _cmd = "wb_command -volume-resample"

    def _format_arg(self, opt, spec, val):
        if opt == "flirt" and val:
            val = (
                self.inputs.flirt_source_volume or self.inputs.in_file,
                self.inputs.flirt_target_volume or self.inputs.volume_space,
            )
        return super()._format_arg(opt, spec, val)


class VolumeAllLabelsToROIsInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=0,
        desc="the input volume label file",
    )
    label_map = traits.Either(
        traits.Int,
        Str,
        mandatory=True,
        argstr="%s",
        position=1,
        desc="the number or name of the label map to use",
    )
    out_file = File(
        name_source=["in_file"],
        name_template="%s_rois.nii.gz",
        argstr="%s",
        position=2,
        desc="the output volume",
    )


class VolumeAllLabelsToROIsOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output volume")


class VolumeAllLabelsToROIs(WBCommand):
    """
    Make ROIs from all labels in a volume frame

    The output volume has a frame for each label in the specified input
    frame, other than the ??? label, each of which contains an ROI of all
    voxels that are set to the corresponding label.

    >>> from nibabies.interfaces.workbench import VolumeAllLabelsToROIs
    >>> rois = VolumeAllLabelsToROIs()
    >>> rois.inputs.in_file = data_dir / 'atlas.nii'
    >>> rois.inputs.label_map = 1
    >>> rois.cmdline
    'wb_command -volume-all-labels-to-rois .../atlas.nii 1 atlas_rois.nii.gz'
    """

    input_spec = VolumeAllLabelsToROIsInputSpec
    output_spec = VolumeAllLabelsToROIsOutputSpec
    _cmd = "wb_command -volume-all-labels-to-rois"


class VolumeLabelExportTableInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=0,
        desc="the input volume label file",
    )
    label_map = traits.Either(
        traits.Int,
        Str,
        mandatory=True,
        argstr="%s",
        position=1,
        desc="the number or name of the label map to use",
    )
    out_file = File(
        name_source=["in_file"],
        name_template="%s_labels.txt",
        argstr="%s",
        position=2,
        desc="the output text file",
    )


class VolumeLabelExportTableOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output text file")


class VolumeLabelExportTable(WBCommand):
    """
    Export label table from volume as text

    Takes the label table from the volume label map, and writes it to a text
    format matching what is expected by -volume-label-import.

    >>> from nibabies.interfaces.workbench import VolumeLabelExportTable
    >>> label_export = VolumeLabelExportTable()
    >>> label_export.inputs.in_file = data_dir / 'atlas.nii'
    >>> label_export.inputs.label_map = 1
    >>> label_export.cmdline
    'wb_command -volume-label-export-table .../atlas.nii 1 atlas_labels.txt'
    """

    input_spec = VolumeLabelExportTableInputSpec
    output_spec = VolumeLabelExportTableOutputSpec
    _cmd = "wb_command -volume-label-export-table"


class VolumeLabelImportInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=0,
        desc="the input volume file",
    )
    label_list_file = File(
        exists=True,
        mandatory=True,
        argstr="%s",
        position=1,
        desc="text file containing the values and names for labels",
    )
    out_file = File(
        name_source=["in_file"],
        name_template="%s_labels.nii.gz",
        argstr="%s",
        position=2,
        desc="the output workbench label volume",
    )
    discard_others = traits.Bool(
        argstr="-discard-others",
        desc="set any voxels with values not mentioned in the label list to the ??? label",
    )
    unlabeled_values = traits.Int(
        argstr="-unlabeled-value %d",
        desc="the value that will be interpreted as unlabeled",
    )
    subvolume = traits.Either(
        traits.Int,
        Str,
        argstr="-subvolume %s",
        desc="select a single subvolume to import (number or name)",
    )
    drop_unused_labels = traits.Bool(
        argstr="-drop-unused-labels",
        desc="remove any unused label values from the label table",
    )


class VolumeLabelImportOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output workbench label volume")


class VolumeLabelImport(WBCommand):
    """
    Import a label volume to workbench format

    Creates a label volume from an integer-valued volume file.  The label
    name and color information is stored in the volume header in a nifti
    extension, with a similar format as in caret5, see -volume-help.  You may
    specify the empty string (use "") for <label-list-file>, which will be
    treated as if it is an empty file.  The label list file must have the
    following format (2 lines per label):

    <labelname>
    <key> <red> <green> <blue> <alpha>
    ...

    Label names are specified on a separate line from their value and color,
    in order to let label names contain spaces.  Whitespace is trimmed from
    both ends of the label name, but is kept if it is in the middle of a
    label.  Do not specify the "unlabeled" key in the file, it is assumed
    that 0 means not labeled unless -unlabeled-value is specified.  The value
    of <key> specifies what value in the imported file should be used as this
    label.  The values of <red>, <green>, <blue> and <alpha> must be integers
    from 0 to 255, and will specify the color the label is drawn as (alpha of
    255 means fully opaque, which is probably what you want).

    By default, it will create new label names with names like LABEL_5 for
    any values encountered that are not mentioned in the list file, specify
    -discard-others to instead set these values to the "unlabeled" key.

    >>> from nibabies.interfaces.workbench import VolumeLabelImport
    >>> label_import = VolumeLabelImport()
    >>> label_import.inputs.in_file = data_dir / 'atlas.nii'
    >>> label_import.inputs.label_list_file = data_dir / 'label_list.txt'
    >>> label_import.cmdline
    'wb_command -volume-label-import .../atlas.nii .../label_list.txt \
    atlas_labels.nii.gz'
    """

    input_spec = VolumeLabelImportInputSpec
    output_spec = VolumeLabelImportOutputSpec
    _cmd = "wb_command -volume-label-import"
